#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import time
from datetime import timedelta, datetime
import click
import random
import sys
import json
import re
from pathlib import Path
import subprocess
from multiprocessing import Pool
import yaml
from box import Box
import ast
import hashlib

from opengate.exception import fatal, colored, color_ok, color_error, color_warning
from opengate_core.testsDataSetup import check_tests_data_folder
from opengate.bin.opengate_library_path import return_tests_path
from opengate_core import GateInfo
from opengate.exception import warning

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("--start_id", "-i", default="all", help="Start test from this number")
@click.option("--end_id", "-e", default="all", help="Start test up to this number")
@click.option(
    "--no_log_on_fail",
    default=False,
    is_flag=True,
    help="If set, do not print log on fail",
)
@click.option(
    "--random_tests",
    "-r",
    is_flag=True,
    default=False,
    help="Start the last 10 tests and 1/4 of the others randomly",
)
@click.option(
    "--seed",
    "-s",
    default="",
    help="Seed for the random generator",
)
@click.option(
    "--processes_run",
    "-p",
    default="mp",
    help="Start simulations in single process mode: 'legacy', 'sp' or multi process mode 'mp'",
)
@click.option(
    "--num_processes",
    "-n",
    default="all",
    help="Start simulations in multiprocessing mode and run n processes - default is all available cores",
)
@click.option(
    "--run_previously_failed_jobs",
    "-f",
    default=False,
    is_flag=True,
    help="Run only the tests that failed in the previous evaluation.",
)
@click.option(
    "--g4_version",
    "-v",
    default="",
    help="Only for developers: overwrite the used geant4 version str to pass the check, style: v11.3.2",
)
def go(
    start_id,
    end_id,
    random_tests,
    seed,
    no_log_on_fail,
    processes_run,
    run_previously_failed_jobs,
    num_processes,
    g4_version,
):

    path_tests_src = return_tests_path()  # returns the path to the tests/src dir
    test_dir_path = path_tests_src.parent
    start = time.time()
    if not g4_version:
        try:
            g4_version = get_required_g4_version(path_tests_src)
        except:
            g4_version = "v11.3.2"
    if not check_g4_version(g4_version):
        warning(f'The geant4 version "{g4_version}" is not the expected version.')
        # return 0

    path_output_dashboard = test_dir_path / "output_dashboard"
    fpath_dashboard_output = path_output_dashboard / (
        "dashboard_output_"
        + sys.platform
        + "_"
        + str(sys.version_info[0])
        + "."
        + str(sys.version_info[1])
        + ".json"
    )

    if not run_previously_failed_jobs:
        files_to_run_avail, files_to_ignore = get_files_to_run()
        files_to_run = select_files(
            files_to_run_avail, start_id, end_id, random_tests, seed
        )
        download_data_at_first_run(files_to_run_avail[0])
        dashboard_dict_out = {k: [""] for k in files_to_run_avail}
    else:
        with open(fpath_dashboard_output, "r") as fp:
            dashboard_dict_out = json.load(fp)
            files_to_run = [k for k, v in dashboard_dict_out.items() if not v[0]]

    files_to_run_part1, files_to_run_part2_depending_on_part1 = (
        filter_files_with_dependencies(files_to_run, path_tests_src)
    )
    # print(f"{' ,'.join(files_to_run_part1)}")
    if len(files_to_run_part2_depending_on_part1) > 0:
        print(
            f"Found test cases with mutual dependencies, going to split evaluation into two sets. {len(files_to_run_part2_depending_on_part1)} tests will start right after first eval round."
        )
    runs_status_info = run_test_cases(
        files_to_run_part1, no_log_on_fail, processes_run, num_processes
    )

    if len(files_to_run_part2_depending_on_part1) > 0:
        print(
            "Now starting evaluation of tests depending on results of previous tests:"
        )
        runs_status_info_part2 = run_test_cases(
            files_to_run_part2_depending_on_part1,
            no_log_on_fail,
            processes_run,
            num_processes,
        )

        dashboard_dict, failure = status_summary_report(
            runs_status_info + runs_status_info_part2,
            files_to_run_part1 + files_to_run_part2_depending_on_part1,
            no_log_on_fail,
        )

    else:
        dashboard_dict, failure = status_summary_report(
            runs_status_info, files_to_run_part1, no_log_on_fail
        )
    end = time.time()
    run_time = timedelta(seconds=end - start)
    print(f"Evaluation took in total:   {run_time.seconds /60 :5.1f} min  ")
    dashboard_dict_out.update(dashboard_dict)
    if fpath_dashboard_output:
        os.makedirs(str(fpath_dashboard_output.parent), exist_ok=True)
        with open(fpath_dashboard_output, "w") as fp:
            json.dump(dashboard_dict_out, fp, indent=4)
    print(not failure)


def get_files_to_run():

    path_tests_src = return_tests_path()
    print("Looking for tests in: " + str(path_tests_src))

    if not check_tests_data_folder():
        return False

    # Look if torch is installed
    torch = True
    torch_tests = [
        "test034_gan_phsp_linac.py",
        "test038_gan_phsp_spect_gan_my.py",
        "test038_gan_phsp_spect_gan_aa.py",
        "test038_gan_phsp_spect_gan_se.py",
        "test038_gan_phsp_spect_gan_ze.py",
        "test038_gan_phsp_spect_gan_mt.py",
        "test040_gan_phsp_pet_gan.py",
        "test043_garf.py",
        "test043_garf_mt.py",
        "test045_speedup_all_wip.py",
        "test047_gan_vox_source_cond.py",
        "test081_simulation_optigan_with_random_seed.py",
        "test085_free_flight_mt.py",
        "test085_free_flight_ref_mt.py",
        "test085_free_flight_rotation.py",
    ]
    try:
        import torch
    except:
        torch = False

    ignored_tests = [
        "test045_speedup",  # this is a binary (still work in progress)
    ]
    path_tests_src = Path(path_tests_src)
    all_file_paths = [
        file.name for file in path_tests_src.glob("test[0-9]*.py") if file.is_file()
    ]
    # here we sort the paths
    all_file_paths = sorted(all_file_paths)

    ignore_files_containing = [
        "wip",
        "visu",
        "old",
        ".log",
        "all_tes",
        "_base",
        "_helpers",
    ]
    ignore_files_containing += ignored_tests
    print(
        f"Going to ignore all file names that contain any of: {', '.join(ignore_files_containing)}"
    )
    files_to_run = []
    files_to_ignore = []

    for filename in all_file_paths:
        eval_this_file = True
        reason_to_ignore = ""
        for string_to_ignore in ignore_files_containing:
            if string_to_ignore.lower() in filename.lower():
                eval_this_file = False
                reason_to_ignore = string_to_ignore
                continue
        if not torch and filename in torch_tests:
            reason_to_ignore = "Torch not avail"
            eval_this_file = False
        if os.name == "nt" and "_mt" in filename:
            eval_this_file = False
            reason_to_ignore = "mt & nt"
        if eval_this_file:
            files_to_run.append(filename)
        else:
            print(
                colored.stylize(
                    f"Ignoring: {filename:<40} --> {reason_to_ignore}",
                    color_warning,
                ),
                end="\n",
            )
            files_to_ignore.append(filename)
    print(
        f"Found {len(all_file_paths)} available test cases, of those {len(files_to_run)} files to run, and {len(files_to_ignore)} ignored."
    )
    return files_to_run, files_to_ignore


def get_required_g4_version(tests_dir: Path, rel_fpath=".github/workflows/main.yml"):
    fpath = tests_dir.parents[2] / rel_fpath
    with open(fpath) as f:
        githubworfklow = yaml.safe_load(f)
    g4 = githubworfklow["jobs"]["build_wheel"]["env"]["GEANT4_VERSION"]
    return g4


def check_g4_version(g4_version: str):
    v = GateInfo.get_G4Version().replace("$Name: ", "")
    v = v.replace("$", "")
    print(f"Detected Geant4 version: {v}")
    print(f"Required Geant4 version: {g4_version}")
    g4_should = decompose_g4_versioning(g4_version)
    g4_is = decompose_g4_versioning(v)
    if g4_should == g4_is:
        print(colored.stylize("Geant4 version is OK", color_ok), end="\n")
        return True
    else:
        print(f'{" ".join(map(str,g4_should))}')
        print(f'{" ".join(map(str,g4_is))}')
        print(colored.stylize("Geant4 version is not ok", color_error), end="\n")
        return False


def decompose_g4_versioning(g4str):
    # Check if patch is present:
    patchedVersion = False
    g4str = g4str.lower()
    if "-patch" in g4str:
        patchedVersion = True
        g4str = g4str.replace("-patch", "")
    # Regular expression pattern to match integers separated by . - _ or p
    pattern = r"\d+(?=[._\-p ])|\d+$"

    # Find all matches
    matches = re.findall(pattern, g4str)
    g4_version = [int(k) for k in matches]
    # removing 4 from "geant4"
    if g4_version and g4_version[0] == int(4):
        g4_version.pop(0)
    if not patchedVersion and len(g4_version) < 3:
        g4_version.append(0)
    return g4_version


def select_files(files_to_run, test_id, end_id, random_tests, seed):
    pattern = re.compile(r"^test([0-9]+)")

    if test_id != "all" or end_id != "all":
        test_id = int(test_id) if test_id != "all" else 0
        end_id = int(end_id) if end_id != "all" else sys.maxsize
        files_new = []
        for f in files_to_run:
            match = pattern.match(f)
            f_test_id = int(float(match.group(1)))
            if f_test_id >= test_id and f_test_id <= end_id:
                files_new.append(f)
            else:
                if f_test_id < test_id:
                    print(f"Ignoring: {f:<40} (< {test_id}) ")
                else:
                    print(f"Ignoring: {f:<40} (> {end_id}) ")
        files_to_run = files_new
    elif random_tests:
        files_new = files_to_run[-10:]
        prob = 0.25
        if not seed == "":
            # Convert string to a hash and then to an integer
            hash_object = hashlib.md5(seed.encode())
            hash_hex = hash_object.hexdigest()
            seed_nb = int(hash_hex, 16) % (2**32)
            random.seed(seed_nb)

        files = files_new + random.sample(
            files_to_run[:-10], int(prob * (len(files_to_run) - 10))
        )
        files_to_run = sorted(files)
    return files_to_run


def get_main_function_args(file_dir, file_path):
    with open(file_dir / file_path, "r") as file:
        tree = ast.parse(file.read(), filename=file_path)

    # Find the 'main' function in the AST
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef) and node.name == "main":
            # Extract the arguments of the main function

            args_info = {}
            # Handle cases where some arguments have no defaults
            num_defaults = len(node.args.defaults)
            num_args = len(node.args.args)
            non_default_args = num_args - num_defaults
            for i, arg in enumerate(node.args.args):
                arg_name = arg.arg
                if i >= non_default_args:
                    # Match argument with its default value
                    default_value = node.args.defaults[i - non_default_args]
                    # default_value_str = ast.dump(default_value).value()
                    if isinstance(default_value, ast.Constant):
                        default_value_str = default_value.value
                    else:
                        default_value_str = None
                else:
                    default_value_str = None
                args_info[arg_name] = default_value_str
            return args_info

    # Return None if no 'main' function is found
    return None


def analyze_scripts(file_dir, files):
    files_dependence_on = {}
    for file_path in files:
        args = get_main_function_args(file_dir, file_path)
        file_depending_on = None
        if args is not None and "dependency" in args:
            file_depending_on = args["dependency"]
        files_dependence_on[file_path] = file_depending_on
    return files_dependence_on


def filter_files_with_dependencies(files_to_run, path_tests_src):
    files_dependence_dict = analyze_scripts(path_tests_src, files_to_run)
    files_needed_for_other_tests = [
        needed_file
        for file, needed_file in files_dependence_dict.items()
        if needed_file
    ]

    files_to_run_part1 = [f for f in files_to_run if not files_dependence_dict[f]]
    files_to_run_part1 += files_needed_for_other_tests
    files_to_run_part1 = list(set(files_to_run_part1))
    files_to_run_part2_depending_on_part1 = [
        f for f in files_to_run if files_dependence_dict[f]
    ]
    return files_to_run_part1, files_to_run_part2_depending_on_part1


def run_one_test_case(f, processes_run, path_tests_src):
    """
    This function is obsolete if we don't need os.system(run_cmd)
    """
    start = time.time()
    print(f"Running: {f:<46}  ", end="")
    cmd = "python " + str(path_tests_src / f)
    log = str(path_tests_src.parent / "log" / f) + ".log"
    if processes_run == "legacy":
        r = os.system(f"{cmd} > {log} 2>&1")
        shell_output = Box({"returncode": r, "log_fpath": log})
    else:
        shell_output = subprocess.run(
            f"{cmd} > {log} 2>&1",
            shell=True,
            check=False,
            capture_output=True,
            text=True,
        )
        r = shell_output.returncode
    if r == 0:
        print(colored.stylize(" OK", color_ok), end="")
    else:
        if r == 2:
            # this is probably a Ctrl+C, so we stop
            fatal("Stopped by user")
        else:
            print(colored.stylize(" FAILED !", color_error), end="")
    end = time.time()
    shell_output.run_time = start - end
    print(f"   {end - start:5.1f} s     {log:<65}")
    return shell_output


def download_data_at_first_run(f):
    print("Running one test case to trigger download of data if not available yet.")
    run_one_test_case_mp(f)


def run_one_test_case_mp(f):
    path_tests_src = return_tests_path()
    start = time.time()
    print(f"Running: {f:<46}  ", end="")
    cmd = "python " + str(path_tests_src / f)
    log = str(path_tests_src.parent / "log" / Path(f).stem) + ".log"

    # Write the command as the first line in the log file
    start = time.time()
    current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    with open(log, "w") as log_file:
        log_file.write(f"\n\nDate/Time: {current_time}\n")
        log_file.write(f"Command: {cmd}\n\n")

    shell_output = subprocess.run(
        f"{cmd} >> {log} 2>&1", shell=True, check=False, capture_output=True, text=True
    )
    shell_output.log_fpath = log
    r = shell_output.returncode
    if r == 0:
        print(colored.stylize(" OK", color_ok), end="")
    else:
        if r == 2:
            # this is probably a Ctrl+C, so we stop
            fatal("Stopped by user")
        else:
            print(colored.stylize(" FAILED !", color_error), end="")
    end = time.time()
    shell_output.run_time = start - end
    print(f"\t {end - start:5.1f} s     {Path(log).name}")
    return shell_output


def run_test_cases(
    files: list, no_log_on_fail: bool, processes_run: str, num_processes: str
):
    path_tests_src = return_tests_path()
    # print("Looking for tests in: " + str(path_tests_src))
    print(f"Running {len(files)} tests")
    print("-" * 70)

    start = time.time()
    if processes_run in ["legacy"]:
        run_single_case = lambda k: run_one_test_case(k, processes_run, path_tests_src)
        runs_status_info = [run_single_case(file) for file in files]
    elif processes_run in ["sp"]:
        runs_status_info = [run_one_test_case_mp(file) for file in files]
    else:
        num_processes = int(float(num_processes)) if num_processes != "all" else None
        with Pool(processes=num_processes) as pool:
            runs_status_info = pool.map(run_one_test_case_mp, files)

    end = time.time()
    run_time = timedelta(seconds=end - start)
    print(f"Running tests took:   {run_time.seconds /60 :5.1f} min  ")
    return runs_status_info


def status_summary_report(runs_status_info, files, no_log_on_fail):

    dashboard_dict = {
        str(Path(k).name): [shell_output_k.returncode == 0]
        for k, shell_output_k in zip(files, runs_status_info)
    }

    tests_passed = [f for f in files if dashboard_dict[f][0]]
    tests_passed.sort()
    tests_failed = [f for f in files if not dashboard_dict[f][0]]
    tests_failed.sort()

    n_passed = sum([k[0] for k in dashboard_dict.values()])
    n_failed = sum([not k[0] for k in dashboard_dict.values()])

    # Display the logs of the failed jobs:
    for file, shell_output_k in zip(files, runs_status_info):
        if shell_output_k.returncode != 0 and not no_log_on_fail:
            print(
                str(Path(file).name),
                colored.stylize(": failed", color_error),
                end="\n",
            )
            if os.name == "nt":
                os.system("type " + shell_output_k.log_fpath)
            else:
                os.system("cat " + shell_output_k.log_fpath)

    print(f"Summary pass: {n_passed}/{len(files)} passed the tests:")
    for k in tests_passed:
        print(str(Path(k).name), colored.stylize(": passed", color_ok), end="\n")

    if n_failed > 0:
        print(f"Summary fail: {n_failed}/{len(files)} failed the tests:")
        for k in tests_failed:
            print(str(Path(k).name), colored.stylize(": failed", color_error), end="\n")

    fail_status = 0
    if n_passed == len(files) and n_failed == 0:
        print(colored.stylize("Yeahh, all tests passed!", color_ok))
        fail_status = 0
    else:
        print(colored.stylize("Oh no, not all tests passed.", color_error))
        fail_status = 1
    return dashboard_dict, fail_status


if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from opengate import g4_units
from opengate.tests import utility
from test085_free_flight_helpers import *

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, None, output_folder="test085_spect"
    )

    # create the simulation
    sim = gate.Simulation()
    sim.number_of_threads = 4
    ac = 1e4
    # sim.visu = True
    source, actors = create_simulation_test085(
        sim,
        paths,
        simu_name="ff_sc",
        ac=ac,
        use_spect_head=True,
        use_spect_arf=False,
        use_phsp=False,
    )

    # free flight actor
    ff = sim.add_actor("ScatterSplittingFreeFlightActor", "ff")
    ff.attached_to = "world"
    ff.unbiased_volumes = ["spect_1_crystal", "spect_2_crystal"]
    ff.kill_interacting_in_volumes = ["spect_1_crystal", "spect_2_crystal"]
    ff.compton_splitting_factor = 50
    ff.rayleigh_splitting_factor = 10
    ff.max_compton_level = 10000
    ff.acceptance_angle.intersection_flag = True
    ff.acceptance_angle.volumes = ["spect_1"]  # FIXME check volume exists before
    ff.acceptance_angle.normal_flag = True
    ff.acceptance_angle.normal_vector = [0, 0, -1]
    ff.acceptance_angle.normal_tolerance = 10 * g4_units.deg

    # go
    sim.run()
    stats = sim.get_actor("stats")
    print(stats)

    print()
    print(ff)

    # compare to noFF
    is_ok = True
    is_ok = (
        utility.assert_images(
            paths.output_ref / "projection_1_ff_sc_counts.mhd",
            paths.output / "projection_1_ff_sc_counts.mhd",
            stats,
            tolerance=300,
            ignore_value_data1=0,
            sum_tolerance=62,
            sad_profile_tolerance=39,
            axis="x",
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.geometry.materials import MaterialDatabase, assert_same_material
from opengate.tests import utility
from scipy.spatial.transform import Rotation

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test009_voxels", "test009")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.output_dir = paths.output

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    mm = gate.g4_units.mm
    gcm3 = gate.g4_units.g_cm3

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # add a simple fake volume to test hierarchy
    # translation and rotation like in the Gate macro
    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.material = "G4_AIR"
    fake.color = [1, 0, 1, 1]
    fake.rotation = Rotation.from_euler("x", -20, degrees=True).as_matrix()

    # image
    patient = sim.add_volume("Image", "patient")
    patient.image = paths.data / "patient-4mm.mhd"
    patient.mother = "fake"
    patient.material = "G4_AIR"  # material used by default
    f1 = str(paths.gate_data / "Schneider2000MaterialsTable.txt")
    f2 = str(paths.gate_data / "Schneider2000DensitiesTable.txt")
    tol = 0.05 * gcm3
    (
        patient.voxel_materials,
        materials,
    ) = gate.geometry.materials.HounsfieldUnit_to_material(sim, tol, f1, f2)
    print(f"tol = {tol/gcm3} g/cm3")
    print(f"mat : {len(patient.voxel_materials)} materials")
    patient.dump_label_image = paths.output / "test009_hu_label.mhd"
    # cuts
    patient.set_production_cut(particle_name="electron", value=3 * mm)

    # dump list of created material (for debug)
    fn = str(paths.output / "test009_materials.txt")
    with open(fn, "w") as file:
        file.write("[Materials]\n")
        for i, m in enumerate(materials):
            file.write(f"# {patient.voxel_materials[i]}\n")
            print(f"Build material number {i} named {m}")
            mat = sim.volume_manager.material_database.FindOrBuildMaterial(m)
            file.write(gate.geometry.materials.dump_material_like_Gate(mat))
    print("List of material in ", fn)

    # test material files
    gate.exception.warning(f"Check materials")
    db1 = MaterialDatabase()
    db1.read_from_file(str(paths.gate_data / "patient-HUmaterials.db"))
    db2 = MaterialDatabase()
    db2.read_from_file(fn)
    is_ok = True
    for m1 in db1.material_builders:
        m2 = db2.material_builders[m1]
        m1 = db1.material_builders[m1]
        t = assert_same_material(m1, m2)
        is_ok = utility.print_test(t, f"check {m1.name}") and is_ok

    # write the image of labels (None by default)

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 130 * MeV
    source.particle = "proton"
    source.position.type = "sphere"
    source.position.radius = 10 * mm
    source.position.translation = [0, 0, -14 * cm]
    source.activity = 10000 * Bq
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test009_hu.mhd"
    dose.attached_to = "patient"
    dose.size = [99, 99, 99]
    dose.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose.output_coordinate_system = "attached_to_image"
    dose.translation = [2 * mm, 3 * mm, -2 * mm]
    dose.hit_type = "random"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = False

    # print info
    print(sim.volume_manager.dump_volumes())

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")

    # start simulation
    sim.run()

    # print results at the end
    gate.exception.warning(f"Check stats")
    print(stats)
    print(dose)

    # tests
    gate.exception.warning(f"Check dose")
    stats_ref = utility.read_stat_file(paths.gate_output / "stat_hu.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.15)
    is_ok = (
        utility.assert_images(
            paths.gate_output / "output_hu-Edep.mhd",
            dose.edep.get_output_path(),
            stats,
            tolerance=35,
            ignore_value_data2=0,
            apply_ignore_mask_to_sum_check=False,  # force legacy behavior
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itk
import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, output_folder="test048_split_spect_projections"
    )
    paths.data = paths.data / "test048_split_spect_projections"

    # input of 4 heads projection images, with 3 energy windows
    input_filenames = []
    for i in range(4):
        input_filenames.append(paths.data / f"ref_projection_{i}.mhd")
    print(f"Reading {len(input_filenames)} images")

    info = gate.image.read_image_info(input_filenames[0])
    print("Image info: ", info.size, info.spacing, info.origin)

    # compute the output images
    nb_ene = 3
    outputs = gate.image.split_spect_projections(input_filenames, nb_ene)

    # write them
    e = 0
    output_filenames = []
    output_filename = str(paths.output / "test048_projection.mhd")
    for o in outputs:
        f = output_filename.replace(".mhd", f"_{e}.mhd")
        itk.imwrite(o, f)
        output_filenames.append(f)
        e += 1

    # ---------------------------------------------------------------
    print()
    gate.exception.warning("Compare image to reference")
    is_ok = True
    for i in range(nb_ene):
        is_ok = (
            utility.assert_images(
                paths.output_ref / f"t048_projection_{i}.mhd",
                output_filenames[i],
                None,
                tolerance=1e-6,
                ignore_value_data2=0,
                axis="x",
            )
            and is_ok
        )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import click
import itk
from opengate.image import split_spect_projections

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("inputs", nargs=-1)
@click.option("--ene", "-e", required=True, help="Number of energy windows")
@click.option("--output", "-o", required=True, help="output filename (.mhd)")
def go(inputs, ene, output):
    """
    Consider as input a set of images from a gate SPECT simulation.
    Each image is assumed to be several slices: one per energy windows times one per angle.
    The different images corresponds to several SPECT heads.

    The created output is on image per energy windows, with all angles (heads + rotations) as slices.
    """

    # compute the projections per energy window
    nb_ene = int(ene)
    outputs = split_spect_projections(inputs, nb_ene)

    # write images
    e = 0
    for o in outputs:
        f = output.replace(".mhd", f"_{e}.mhd")
        itk.imwrite(o, f)
        e += 1


# --------------------------------------------------------------------------
if __name__ == "__main__":
    go()



import opengate as gate
from scipy.spatial.transform import Rotation
import numpy as np
from opengate.utility import LazyModuleLoader

sp = LazyModuleLoader("spekpy")
import pathlib

# useful units
MeV = gate.g4_units.MeV
keV = gate.g4_units.keV
Bq = gate.g4_units.Bq
deg = gate.g4_units.deg
nm = gate.g4_units.nm
mm = gate.g4_units.mm
m = gate.g4_units.m
cm = gate.g4_units.cm

current_path = pathlib.Path(__file__).parent.resolve()


class Ciosalpha:
    def __init__(self, sim, kvp, source_only=False):
        self.sim = sim
        self.machine_name = "ciosalpha"
        self.source_only = source_only
        self.volume = self.add_carm_box()
        self.add_xray_tank()
        self.add_collimators()
        self.source = self.add_carm_source(kvp)

    def add_carm_box(self):
        carmbox = self.sim.volume_manager.create_volume("Box", "carmbox")
        carmbox.material = "G4_AIR"
        carmbox.size = [200 * cm, 10 * cm, 120 * cm]
        carmbox.translation = [0 * cm, 0 * cm, 0 * cm]
        carmbox.color = [1, 1, 1, 0.8]

        hole1 = self.sim.volume_manager.create_volume("Box", "hole")
        hole1.size = [191 * cm, 31 * cm, 80 * cm]
        hole1.color = [1, 1, 1, 0.8]
        hole2 = self.sim.volume_manager.create_volume("Box", "hole2")
        hole2.size = [101 * cm, 31 * cm, 31 * cm]
        hole2.color = [1, 1, 1, 0.8]
        hole3 = self.sim.volume_manager.create_volume("Box", "hole3")
        hole3.size = [90 * cm, 31 * cm, 31 * cm]
        hole3.color = [1, 1, 1, 0.8]

        hole1and2 = gate.geometry.volumes.unite_volumes(
            hole1, hole2, translation=[-50.5 * cm, 0 * cm, 55 * cm]
        )

        subtract_to_carm = gate.geometry.volumes.unite_volumes(
            hole1and2, hole3, translation=[-50.5 * cm, 0 * cm, -55 * cm]
        )

        if self.source_only:
            hole4 = self.sim.volume_manager.create_volume("Box", "hole4")
            hole4.size = [45 * cm, 31 * cm, 31 * cm]
            hole4.color = [1, 1, 1, 0.8]
            hole5 = self.sim.volume_manager.create_volume("Box", "hole5")
            hole5.size = [100 * cm, 31 * cm, 121 * cm]
            hole5.color = [1, 1, 1, 0.8]

            hole4and5 = gate.geometry.volumes.unite_volumes(
                hole4, hole5, translation=[55 * cm, 0 * cm, 45 * cm]
            )

            subtract_to_carm = gate.geometry.volumes.unite_volumes(
                subtract_to_carm, hole4and5, translation=[5 * cm, 0 * cm, -35 * cm]
            )

        carm = gate.geometry.volumes.subtract_volumes(
            carmbox,
            subtract_to_carm,
            new_name=self.machine_name,
            translation=[-5 * cm, 0 * cm, -10 * cm],
        )
        self.sim.add_volume(carm)

        return carm

    def add_xray_tank(self):
        xray_tank = self.sim.add_volume("Box", f"{self.machine_name}_xray_tank")
        xray_tank.mother = self.machine_name
        xray_tank.material = "G4_AIR"
        xray_tank.size = [10 * cm, 10 * cm, 30 * cm]
        xray_tank.translation = [0 * cm, 0, 45 * cm]
        xray_tank.color = [1, 1, 1, 0.8]

    def add_carm_source(self, kvp):
        s = sp.Spek(kvp, th=10, physics="kqp")
        s.filter("Al", 3.0).filter("Cu", 0.1)

        energy_bins, weights = s.get_spectrum()

        sourcebox = self.sim.add_volume("Box", f"{self.machine_name}_sourcebox")
        sourcebox.mother = f"{self.machine_name}_xray_tank"
        sourcebox.translation = [0 * cm, 0 * cm, 10 * cm]
        sourcebox.size = [1 * cm, 1 * cm, 1 * cm]
        sourcebox.rotation = Rotation.from_euler(
            "ZYX", [0, 90, 90], degrees=True
        ).as_matrix()

        source = self.sim.add_source("GenericSource", f"{self.machine_name}_source")
        source.attached_to = sourcebox.name
        source.particle = "gamma"
        source.position.type = "disc"
        source.position.radius = 0 * mm

        source.direction_relative_to_attached_volume = True
        source.direction.type = "histogram"
        source.direction.histogram_theta_weights = [1]
        source.direction.histogram_theta_angles = [85 * deg, 95 * deg]

        # TODO: Need real values for the anode heel effect
        data = np.load(current_path / "anodeheeleffect.npz")
        source.direction.histogram_phi_weights = data["weight"]
        source.direction.histogram_phi_angles = data["angle"]

        source.energy.type = "histogram"
        source.energy.histogram_weight = weights
        source.energy.histogram_energy = energy_bins

        source.direction.acceptance_angle.volumes = [sourcebox.name]
        source.direction.acceptance_angle.normal_flag = True
        source.direction.acceptance_angle.normal_vector = [1, 0, 0]
        source.direction.acceptance_angle.normal_tolerance = 5 * deg
        source.direction.acceptance_angle.skip_policy = "SkipEvents"

        return source

    def add_collimators(self):
        xray_tank = self.sim.volume_manager.get_volume(f"{self.machine_name}_xray_tank")
        z_xray_tank = xray_tank.size[2]

        collimators = [
            {
                "translation": [37.5 * mm, 0 * cm, -z_xray_tank / 2 * mm + 1 * mm],
                "size": [2.5 * cm, 5 * cm, 1 * mm],
            },
            {
                "translation": [-37.5 * mm, 0 * cm, -z_xray_tank / 2 * mm + 1 * mm],
                "size": [2.5 * cm, 5 * cm, 1 * mm],
            },
            {
                "translation": [0 * cm, 37.5 * mm, -z_xray_tank / 2 * mm + 3 * mm],
                "size": [5 * cm, 2.5 * cm, 1 * mm],
            },
            {
                "translation": [0 * cm, -37.5 * mm, -z_xray_tank / 2 * mm + 3 * mm],
                "size": [5 * cm, 2.5 * cm, 1 * mm],
            },
        ]

        for i, colli in enumerate(collimators):
            collimator = self.sim.add_volume(
                "Box", f"{self.machine_name}_collimator{i+1}"
            )
            collimator.mother = f"{self.machine_name}_xray_tank"
            collimator.color = [1, 0.7, 0.7, 0.8]
            collimator.translation = colli["translation"]
            collimator.size = colli["size"]

        killer = self.sim.add_actor("KillActor", f"target_kill")
        killer.attached_to = [f"{self.machine_name}_collimator{i+1}" for i in range(4)]

    def set_collimation(self, collimation1, collimation2):
        if not 0 <= collimation1 <= 25 or not 0 <= collimation2 <= 25:
            raise ValueError("Collimation values must be between 0 and 25 mm")

        collimation1 = 25 - collimation1
        collimation2 = 25 - collimation2

        xray_tank = self.sim.volume_manager.get_volume(f"{self.machine_name}_xray_tank")
        z_xray_tank = xray_tank.size[2]

        translations = [
            [37.5 * mm - collimation1, 0 * cm, -z_xray_tank / 2 * mm + 1 * mm],
            [-37.5 * mm + collimation1, 0 * cm, -z_xray_tank / 2 * mm + 1 * mm],
            [0 * cm, 37.5 * mm - collimation2, -z_xray_tank / 2 * mm + 3 * mm],
            [0 * cm, -37.5 * mm + collimation2, -z_xray_tank / 2 * mm + 3 * mm],
        ]

        for i, translation in enumerate(translations):
            collimator = self.sim.volume_manager.get_volume(
                f"{self.machine_name}_collimator{i+1}"
            )
            collimator.translation = translation

    @property
    def collimation(self):
        return self._collimation

    @collimation.setter
    def collimation(self, value):
        self._collimation = value
        self.set_collimation(value[0], value[1])

    @property
    def rotation(self):
        return self.volume.rotation

    @rotation.setter
    def rotation(self, value):
        self.volume.rotation = value

    @property
    def translation(self):
        return self.volume.translation

    @translation.setter
    def translation(self, value):
        self.volume.translation = value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import opengate as gate
import opengate.contrib.phantoms.nemaiec as gate_iec
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test032")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.number_of_threads = 1
    sim.check_volumes_overlap = True
    sim.random_seed = 1548765
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    Bq = gate.g4_units.Bq
    kBq = Bq * 1000

    # world
    sim.world.size = [3 * m, 3 * m, 3 * m]

    # add a first iec phantom (analytical)
    iec1 = gate_iec.add_iec_phantom(sim, "iec1")
    iec1.translation = [40 * cm, 0 * cm, 0 * cm]
    # rotation should have no effect
    iec1.rotation = Rotation.from_euler("y", 33, degrees=True).as_matrix()

    # to highlight the position, we change some volume with high density lead
    v = sim.volume_manager.volumes["iec1_sphere_37mm"]
    v.material = "G4_LEAD_OXIDE"
    v = sim.volume_manager.volumes["iec1_center_cylinder_hole"]
    v.material = "G4_LEAD_OXIDE"

    # add a second iec phantom (voxelized)
    iec2 = sim.add_volume("Image", "iec2")
    iec2.image = paths.output_ref / "test032_iec.mhd"
    iec2.material = "G4_AIR"
    iec2.translation = [-40 * cm, 0 * cm, 0 * cm]
    iec2.dump_label_image = paths.output / "test032_iec_label.mhd"
    labels = utility.read_json_file(paths.output_ref / "test032_labels.json")
    iec2.voxel_materials = []
    for l in labels:
        mat = "IEC_PLASTIC"
        if "capillary" in l:
            mat = "G4_WATER"
        if "cylinder_hole" in l:
            mat = "G4_LUNG_ICRP"
        if "interior" in l:
            mat = "G4_WATER"
        if "sphere" in l:
            mat = "G4_WATER"
        if "world" in l:
            mat = "G4_AIR"
        if "shell" in l:
            mat = "IEC_PLASTIC"
        if "sphere_37mm" in l:
            mat = "G4_LEAD_OXIDE"
        if "center_cylinder_hole" in l:
            mat = "G4_LEAD_OXIDE"
        m = [labels[l]["label"], labels[l]["label"] + 1, mat]
        iec2.voxel_materials.append(m)

    pMin, pMax = sim.volume_manager.volumes["iec1"].bounding_limits
    print("pMin and pMax of iec1", pMin, pMax)

    # the origin of iec1 is different from the origin of iec2
    # we create fake images to be able to convert from
    # the image coordinate space to iec1 or iec2
    # Coordinate system of iec1 is pMin (the extend)
    # Coordinate system of iec2 is the center of the image bounding box
    iec2.load_input_image()
    fake1 = gate.image.create_image_like(iec2.itk_image)
    pMin = gate.geometry.utility.vec_g4_as_np(pMin)
    fake1.SetOrigin(pMin)

    fake2 = gate.image.create_image_like(iec2.itk_image)
    info = gate.image.get_info_from_image(fake2)
    origin = -info.size * info.spacing / 2.0 + info.spacing / 2.0
    fake2.SetOrigin(origin)

    # sources
    activity = 10 * kBq
    for i in range(1, 3):
        source = sim.add_source("GenericSource", f"src{i}")
        source.attached_to = f"iec{i}"
        source.energy.mono = 100 * MeV
        source.particle = "proton"
        source.position.type = "sphere"
        source.position.radius = 10 * mm
        # WARNING the center of the volume is different in the image (iec2)
        # and in the analytical phantom (iec1)
        p = [31 * mm, 33 * mm, 36 * mm]
        if i == 1:
            p = gate.image.transform_images_point(p, iec2.itk_image, fake1)
        else:
            p = gate.image.transform_images_point(p, iec2.itk_image, fake2)
        source.position.translation = p
        source.activity = activity
        source.direction.type = "iso"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # add dose actor
    for i in range(1, 3):
        dose = sim.add_actor("DoseActor", f"dose{i}")
        dose.output_filename = f"test032_iec{i}.mhd"
        dose.attached_to = f"iec{i}"
        dose.size = [100, 100, 100]
        dose.spacing = [2 * mm, 2 * mm, 2 * mm]
        # translate the iec1 to have the exact same dose origin
        # (only needed to perform the assert_image test)
        if i == 1:
            dose.translation = [0 * mm, 35 * mm, 0 * mm]
        if i == 2:
            # only for voxelized:
            dose.output_coordinate_system = "attached_to_image"

    # initialize & start
    sim.run()

    # stats
    print(stats)
    dose1 = sim.get_actor("dose1")
    dose2 = sim.get_actor("dose2")
    # compare edep map

    is_ok = utility.assert_images(
        dose1.edep.get_output_path(),
        dose2.edep.get_output_path(),
        stats,
        tolerance=87,
        axis="x",
        ignore_value_data2=0,
        apply_ignore_mask_to_sum_check=False,  # reproduce legacy behavior of assert_images()
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", output_folder="test084")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    Bq = gate.g4_units.Bq
    mm = gate.g4_units.mm

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # image
    patient = sim.add_volume("Image", "patient")
    patient.image = paths.data / "patient-4mm.mhd"
    patient.material = "G4_AIR"  # material used by default
    patient.voxel_materials = [
        [-2000, -900, "G4_AIR"],
        [-900, -100, "Lung"],
        [-100, 0, "G4_ADIPOSE_TISSUE_ICRP"],
        [0, 300, "G4_TISSUE_SOFT_ICRP"],
        [300, 800, "G4_B-100_BONE"],
        [800, 6000, "G4_BONE_COMPACT_ICRU"],
    ]

    # mu map actor (process at the first begin of run only)
    mumap = sim.add_actor("AttenuationImageActor", "mumap")
    mumap.image_volume = patient  # FIXME volume for the moment, not the name
    mumap.output_filename = "mumap.mhd"
    mumap.energy = 140.511 * keV
    mumap.database = "NIST"  # EPDL
    mumap.attenuation_image.write_to_disk = True
    mumap.attenuation_image.active = True
    print(f"Energy is {mumap.energy/keV} keV")
    print(f"Database is {mumap.database}")

    # go
    sim.run()

    # compare with ref
    ref = paths.output_ref / mumap.output_filename
    is_ok = utility.assert_images(
        ref,
        mumap.get_output_path(),
        tolerance=1e-4,
        fig_name=paths.output / "mumap1.png",
        sum_tolerance=1e-4,
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test008_dose_actor", "test041"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 123456
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq

    #  change world size
    world = sim.world
    world.size = [0.5 * m, 0.5 * m, 0.5 * m]

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [10 * cm, 10 * cm, 10 * cm]
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]

    # lungbox
    lungbox = sim.add_volume("Box", "lungbox")
    lungbox.mother = waterbox.name
    lungbox.size = [10 * cm, 10 * cm, 4 * cm]
    lungbox.translation = [0 * cm, 0 * cm, 2.5 * cm]
    lungbox.material = "G4_LUNG_ICRP"
    lungbox.color = [0, 1, 1, 1]

    # bonebox
    bonebox = sim.add_volume("Box", "bonebox")
    bonebox.mother = waterbox.name
    bonebox.size = [10 * cm, 10 * cm, 4 * cm]
    bonebox.translation = [0 * cm, 0 * cm, -2.5 * cm]
    bonebox.material = "G4_BONE_CORTICAL_ICRP"
    bonebox.color = [1, 0, 0, 1]

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMV"
    sim.physics_manager.global_production_cuts.all = 1 * mm

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 115 * MeV
    source.particle = "proton"
    source.position.type = "disc"
    source.position.radius = 1 * cm
    source.position.translation = [0, 0, -80 * mm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 5000 * Bq

    # add dose actor
    dose_actor = sim.add_actor("DoseActor", "dose_actor")
    # let the actor score other quantities additional to edep (default)
    dose_actor.edep_uncertainty.active = True
    dose_actor.edep_squared.active = True
    dose_actor.dose.active = True
    dose_actor.dose_uncertainty.active = True
    dose_actor.dose_squared.active = True
    # set the filename once for the actor
    # a suffix will be added automatically for each output,
    # i.e. _edep, _edep_uncertainty, _dose
    dose_actor.output_filename = "test041.mhd"
    dose_actor.attached_to = waterbox
    dose_actor.size = [10, 10, 50]
    mm = gate.g4_units.mm
    ts = [200 * mm, 200 * mm, 200 * mm]
    dose_actor.spacing = [x / y for x, y in zip(ts, dose_actor.size)]
    print(dose_actor.spacing)
    dose_actor.hit_type = "random"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # start simulation
    sim.run(start_new_process=True)

    # print results at the end
    print(stats)
    print(dose_actor)

    # tests
    gate.exception.warning("Tests stats file")
    stats_ref = utility.read_stat_file(paths.gate_output / "stat2.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.10)

    gate.exception.warning("\nDifference for EDEP")
    is_ok = (
        utility.assert_images(
            paths.gate_output / "output2-Edep.mhd",
            dose_actor.edep.get_output_path(),
            stats,
            tolerance=10,
            ignore_value_data2=0,
        )
        and is_ok
    )

    gate.exception.warning("\nDifference for EDEP uncertainty")
    is_ok = (
        utility.assert_images(
            paths.gate_output / "output2-Edep-Uncertainty.mhd",
            dose_actor.edep_uncertainty.get_output_path(),
            stats,
            tolerance=30,
            ignore_value_data2=0,
        )
        and is_ok
    )

    gate.exception.warning("\nDifference for EDEP square")
    is_ok = (
        utility.assert_images(
            paths.gate_output / "output2-Edep-Squared.mhd",
            dose_actor.edep_squared.get_output_path(),
            stats,
            tolerance=10,
            ignore_value_data2=0,
        )
        and is_ok
    )

    gate.exception.warning("\nDifference for dose in Gray")
    is_ok = (
        utility.assert_images(
            paths.gate_output / "output2-Dose.mhd",
            dose_actor.dose.get_output_path(),
            stats,
            tolerance=10,
            ignore_value_data2=0,
        )
        and is_ok
    )

    gate.exception.warning("\nDifference for dose uncertainty")
    is_ok = (
        utility.assert_images(
            paths.gate_output / "output2-Dose-Uncertainty.mhd",
            dose_actor.dose_uncertainty.get_output_path(),
            stats,
            tolerance=30,
            ignore_value_data2=0,
        )
        and is_ok
    )

    gate.exception.warning("\nDifference for Does square")
    is_ok = (
        utility.assert_images(
            paths.gate_output / "output2-Dose-Squared.mhd",
            dose_actor.dose_squared.get_output_path(),
            stats,
            tolerance=10,
            ignore_value_data2=0,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate.tests.utility as utility
import numpy as np
import opengate_core as g4
from box import Box
from scipy.spatial.transform import Rotation

if __name__ == "__main__":
    v = g4.G4ThreeVector(2)
    print("v =", v)
    assert v == g4.G4ThreeVector(2, 0, 0)

    print("v.x =", v.x)
    v.x = 1
    v.y = 5
    v.z = 4
    print("v=", v)
    assert v == g4.G4ThreeVector(1, 5, 4)

    print(f"v mag = {v.mag()}")
    np.testing.assert_almost_equal(v.mag(), 6.48074069840786)

    w = g4.G4ThreeVector(v)
    print(f"w = {w}")
    assert v == w
    w.x = 6
    w.y = 5
    w.z = 4
    print(f"w = {w}")
    assert w == g4.G4ThreeVector(6, 5, 4)
    assert v == g4.G4ThreeVector(1, 5, 4)

    print(f"2*v = {2 * v}")
    assert 2 * v == g4.G4ThreeVector(2, 10, 8)

    print(f"v.w = {v * w}")
    assert v * w == 47.0

    v -= w
    print(f"v -= w -> {v}")
    assert v == g4.G4ThreeVector(-5, 0, 0)

    v[0] = 666
    print(v)
    try:
        v[40] = 12
        print(v)
        print(v[12])
        print(v[40])
    except:
        print("Index error ok")

    print("tol", g4.G4ThreeVector.getTolerance())

    # overloaded
    # help(v.isNear)
    print("is near", v.isNear(w, 20))
    assert v.isNear(w, 20) == True

    print("is near", v.isNear(w))
    assert v.isNear(w) == False

    # compare numpy matrix copied to G4.
    u = Box()
    r1 = Rotation.from_euler("y", 23, degrees=True)
    r2 = Rotation.from_euler("x", 66, degrees=True)
    u.mat = (r2 * r1).as_matrix()
    g4_m = g4.DictGetG4RotationMatrix(u, "mat")

    is_ok = True

    def compare(a, b):
        for i in range(3):
            if not np.isclose(a[i], b[i]):
                print("Matrices are not egal")
                print(u.mat)
                print(g4_m)
                return False
            return True

    print("np matrix", u.mat)
    print("G4 matrix", g4_m)
    is_ok = compare(u.mat[0], g4_m.rowX()) and is_ok
    is_ok = compare(u.mat[1], g4_m.rowY()) and is_ok
    is_ok = compare(u.mat[2], g4_m.rowZ()) and is_ok
    utility.print_test(is_ok, "Matrix comparison")

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.utility import g4_units
import opengate.tests.utility as utility
import opengate_core as g4
import math

if __name__ == "__main__":
    # Get Nist Material Manager
    n = g4.G4NistManager.Instance()
    print("Nist manager", n)

    # http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Appendix/materialNames.html
    mat = n.FindOrBuildMaterial("G4_WATER")
    print("Mat Water", mat)
    assert mat.GetName() == "G4_WATER"
    gcm3 = g4_units.g_cm3
    print("Density ", mat.GetDensity(), mat.GetDensity() / gcm3)
    assert math.isclose(mat.GetDensity(), gcm3)
    print("Elements", mat.GetElementVector())
    print("Nb of elements", mat.GetNumberOfElements())
    assert mat.GetNumberOfElements() == 2
    elements = mat.GetElementVector()
    assert elements[0].GetSymbol() == "H"
    assert elements[1].GetSymbol() == "O"

    eV = g4_units.eV
    Imean = mat.GetIonisation().GetMeanExcitationEnergy()
    print("I mean = ", Imean / eV, "eV")
    assert math.isclose(Imean / eV, 78.0)

    # Another material
    mat = n.FindOrBuildMaterial("G4_TISSUE-PROPANE")
    print("Mat 2", mat, mat.GetName())
    N = mat.GetElementVector()[2]
    gmol = g4_units.g_mol
    print("N Z", N.GetZ())
    print("N A", N.GetA() / gmol)
    assert N.GetZ() == 7
    assert math.isclose(N.GetA() / gmol, 14.00676896)

    # n.ListMaterials('all')

    # simple simulation object
    print("-" * 80)
    sim = gate.Simulation()
    print(sim)

    utility.test_ok(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import itk
from scipy.spatial.transform import Rotation

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.number_of_threads = 1
    print(sim)

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    keV = gate.g4_units.keV
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    #  change world size
    sim.world.size = [1.5 * m, 1 * m, 1 * m]

    # fake box
    b = sim.add_volume("Box", "fake1")
    b.size = [36 * cm, 36 * cm, 36 * cm]
    b.translation = [25 * cm, 0, 0]
    r = Rotation.from_euler("y", -25, degrees=True)
    r = r * Rotation.from_euler("x", -35, degrees=True)
    b.rotation = r.as_matrix()
    b = sim.add_volume("Box", "fake2")
    b.mother = "fake1"
    b.size = [35 * cm, 35 * cm, 35 * cm]

    # CT image #1
    ct_odd = sim.add_volume("Image", "ct_odd")
    ct_odd.image = paths.data / "10x10x10.mhd"
    ct_odd.mother = "fake2"
    ct_odd.voxel_materials = [[0, 10, "G4_WATER"]]
    ct_odd.translation = [-2 * cm, 0, 0]
    r = Rotation.from_euler("y", 25, degrees=True)
    r = r * Rotation.from_euler("x", 35, degrees=True)
    ct_odd.rotation = r.as_matrix()

    # CT image #2
    ct_even = sim.add_volume("Image", "ct_even")
    ct_even.image = paths.data / "11x11x11.mhd"
    ct_even.voxel_materials = [[0, 10, "G4_WATER"]]
    ct_even.voxel_materials = ct_odd.voxel_materials
    ct_even.translation = [-25 * cm, 0, 0]

    # source from sphere
    """
        WARNING : if the source is a point and is centered with odd image, the source
        is at the intersection of 3 planes (8 voxels): then, lot of "navigation warning"
        from G4 occur. Not really clear why.
        So we move the source a bit.
    """
    source = sim.add_source("GenericSource", "s_odd")
    source.particle = "alpha"
    source.activity = 1000 * Bq / sim.number_of_threads
    source.direction.type = "iso"
    source.attached_to = "ct_odd"
    source.position.translation = [10 * mm, 10 * mm, 10 * mm]
    source.energy.mono = 1 * MeV

    # source from sphere
    source = sim.add_source("GenericSource", "s_even")
    source.particle = "alpha"
    source.activity = 1000 * Bq / sim.number_of_threads
    source.direction.type = "iso"
    source.attached_to = "ct_even"
    source.position.translation = [0 * mm, 0 * mm, 0 * mm]
    source.energy.mono = 1 * MeV

    # source from spect
    source = sim.add_source("VoxelSource", "vox")
    source.attached_to = "ct_even"
    source.particle = "alpha"
    source.activity = 1000 * Bq / sim.number_of_threads
    source.image = paths.data / "five_pixels.mha"
    source.direction.type = "iso"
    source.position.translation = [0 * mm, 0 * mm, 0 * mm]
    source.position.translation = gate.image.get_translation_between_images_center(
        str(ct_even.image), str(source.image)
    )
    source.energy.mono = 1 * MeV

    # cuts
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMZ"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.all = 1 * mm

    # add dose actor
    dose1 = sim.add_actor("DoseActor", "dose1")
    dose1.output = paths.output / "test021-odd-edep.mhd"
    dose1.mother = "ct_odd"
    img_info = gate.image.read_image_info(str(ct_odd.image))
    dose1.size = img_info.size
    dose1.spacing = img_info.spacing
    dose1.img_coord_system = True

    # add dose actor
    dose2 = sim.add_actor("DoseActor", "dose2")
    dose2.output = paths.output / "test021-even-edep.mhd"
    dose2.mother = "ct_even"
    img_info = gate.image.read_image_info(str(ct_even.image))
    dose2.size = img_info.size
    dose2.spacing = img_info.spacing
    dose2.translation = source.position.translation
    dose2.img_coord_system = True

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")

    # start simulation
    sim.run()

    # print results at the end
    stat = sim.get_actor("Stats")
    # stat.write('output_ref/stat021_ref.txt')

    # test pixels in dose #1
    dose1 = sim.get_actor("dose1")
    dose2 = sim.get_actor("dose2")
    d_odd = itk.imread(paths.output / dose1.user_info.output)
    s = itk.array_view_from_image(d_odd).sum()
    v = d_odd.GetPixel([5, 5, 5])
    diff = (s - v) / s
    tol = 0.01
    is_ok = diff < tol
    diff *= 100
    utility.print_test(is_ok, f"Image #1 (odd): {v:.2f} {s:.2f} -> {diff:.2f}%")

    # test pixels in dose #1
    d_even = itk.imread(paths.output / dose2.user_info.output)
    s = itk.array_view_from_image(d_even).sum()
    v0 = d_even.GetPixel([5, 5, 5])
    v1 = d_even.GetPixel([1, 5, 5])
    v2 = d_even.GetPixel([1, 2, 5])
    v3 = d_even.GetPixel([5, 2, 5])
    v4 = d_even.GetPixel([6, 2, 5])
    tol = 0.15
    ss = v0 + v1 + v2 + v3 + v4

    def t(s, v):
        diff = abs(s - v) / s
        b = diff < tol
        p = diff * 100.0
        utility.print_test(b, f"Image #2 (even) {s:.2f} vs {v:.2f}  -> {p:.2f}%")
        return b

    is_ok = t(s, ss) and is_ok
    is_ok = t(1.80, v0) and is_ok
    is_ok = t(0.8, v1) and is_ok
    is_ok = t(0.8, v2) and is_ok
    is_ok = t(0.8, v3) and is_ok
    is_ok = t(0.8, v4) and is_ok

    stats_ref = utility.read_stat_file(paths.output_ref / "stat021_ref.txt")
    stats_ref.counts.runs = sim.number_of_threads
    is_ok = utility.assert_stats(stat, stats_ref, 0.05) and is_ok

    utility.test_ok(is_ok)



import opengate.engines
import opengate_core as g4

import opengate.physics
from opengate.utility import get_material_name_variants


def get_ions_generated_per_spot(simulation_engine):
    sources = [
        s
        for s in simulation_engine.source_engine.sources
        if s.type_name == "TreatmentPlanPBSource"
    ]
    generated_primaries = {}
    for i, s in enumerate(sources):
        print(s.user_info.name)
        generated_primaries[s.user_info.name + f"_{i}"] = s.get_generated_primaries()
    print(generated_primaries)


def check_production_cuts(simulation_engine):
    """Function to be called by opengate after initialization
    of the simulation, i.e. when G4 volumes and regions exist.
    The purpose is to check whether Geant4 has properly set
    the production cuts in the specific region.

    The value max_step_size is stored in the attribute hook_log
    which can be accessed via the output of the simulation.

    """
    print(f"Entered hook")
    rs = g4.G4RegionStore.GetInstance()
    print("Known regions are:")
    for i in range(rs.size()):
        print("*****")
        print(f"{rs.Get(i).GetName()}")
        reg = rs.Get(i)
        pcuts = reg.GetProductionCuts()
        if pcuts is not None:
            cut_proton = pcuts.GetProductionCut("proton")
            cut_positron = pcuts.GetProductionCut("e+")
            cut_electron = pcuts.GetProductionCut("e-")
            cut_gamma = pcuts.GetProductionCut("gamma")
            print("Cuts in this region:")
            print(f"gamma: {cut_gamma}")
            print(f"electron: {cut_electron}")
            print(f"proton: {cut_proton}")
            print(f"positron: {cut_positron}")
        else:
            print("Found no cuts in this region")


def user_hook_dump_material_properties(simulation_engine):
    print("*** In user hook dump_material_properties ***")
    for vol in simulation_engine.simulation.volume_manager.volumes.values():
        material_name = vol.g4_material.GetName()
        material_dict = opengate.physics.load_optical_properties_from_xml(
            simulation_engine.simulation.physics_manager.optical_properties_file,
            material_name,
        )
        print(f"Volume {vol.name} has material {material_name}")
        mpt = vol.g4_material.GetMaterialPropertiesTable()
        if mpt is not None and material_dict is not None:
            const_prop_names = mpt.GetMaterialConstPropertyNames()
            vector_prop_names = mpt.GetMaterialPropertyNames()
            if not set(material_dict["constant_properties"].keys()).issubset(
                set([str(n) for n in const_prop_names])
            ):
                print(
                    "NOT all constant_properties from file found in G4MaterialPropertiesTable"
                )
                simulation_engine.user_hook_log.append(False)
            else:
                simulation_engine.user_hook_log.append(True)
            if not set(material_dict["vector_properties"].keys()).issubset(
                set([str(n) for n in vector_prop_names])
            ):
                print(
                    "NOT all vector_properties from file found in G4MaterialPropertiesTable"
                )
                simulation_engine.user_hook_log.append(False)
            else:
                simulation_engine.user_hook_log.append(True)
        elif mpt is None and material_dict is not None:
            print(
                f"Geant4 does not find any MaterialPropertiesTable for this material "
                f"although it is defined in the optical_properties_file "
                f"{simulation_engine.simulation.physics_manager.optical_properties_file}"
            )
            simulation_engine.user_hook_log.extend([False, False])
    print("*** ------------------------------------- ***")


def user_hook_em_switches(simulation_engine):
    switches = {}
    switches["auger"] = simulation_engine.physics_engine.g4_em_parameters.Auger()
    switches["fluo"] = simulation_engine.physics_engine.g4_em_parameters.Fluo()
    switches["pixe"] = simulation_engine.physics_engine.g4_em_parameters.Pixe()
    switches["auger_cascade"] = (
        simulation_engine.physics_engine.g4_em_parameters.AugerCascade()
    )
    switches["deexcitation_ignore_cut"] = (
        simulation_engine.physics_engine.g4_em_parameters.DeexcitationIgnoreCut()
    )
    simulation_engine.user_hook_log.append(switches)
    print("Found the following em parameters via the user hook:")
    for k, v in switches.items():
        print(f"{k}: {v}")


def user_hook_active_regions(simulation_engine):
    active_regions = {}
    active_regions["world"] = g4.check_active_region("DefaultRegionForTheWorld")
    active_regions["world"] = g4.check_active_region("DefaultRegionForTheWorld")
    for region in simulation_engine.simulation.physics_manager.regions.values():
        active_regions[region.name] = g4.check_active_region(region.name)
    print(f"Found the following em switches via the user hook:")
    for r, s in active_regions.items():
        print(f"Region {r}:")
        print(f"    deexcitation activated: {s[0]}")
        print(f"    auger activated: {s[1]}")
    simulation_engine.user_hook_log.append(active_regions)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test053_phid_helpers1 import *
import opengate as gate

if __name__ == "__main__":
    """
    Consider a source of fr221 and store all emitted gammas
    """

    """
    WARNING
    PhotonIonDecayIsomericTransitionExtractor does NOT work anymore
    Now PHID extract data from the IAEA database, not directly from G4
    """
    fatal("PhotonIonDecayIsomericTransitionExtractor NOT implemented")

    test_ok(False)

    paths = get_default_test_paths(__file__, "", output_folder="test053")
    z = 87
    a = 221
    sim = gate.Simulation()

    ion_name, daughters = create_ion_gamma_simulation(sim, paths, z, a)

    # go
    sim.run(start_new_process=True)

    #
    is_ok = analyse(
        paths, sim, sim.output, ion_name, z, a, daughters, log_flag=False, tol=0.08
    )

    test_ok(is_ok)



import os
from setuptools import setup

with open("VERSION", "r") as fh:
    version = fh.read()[:-1]

# For windows, this package is needed
install_requires_windows = []
if os.name == "nt":
    install_requires_windows = ["msvc-runtime"]

setup(
    install_requires=[
        "colored>1.5",
        "opengate-core==" + version,
        "gatetools",
        "click",
        "python-box<7.0.0",
        "anytree",
        "numpy",
        "itk",
        "uproot",
        "scipy",
        "matplotlib",
        "GitPython",
        "meshio",
        "radioactivedecay",
        "jsonpickle",
        "pandas",
        "awkward-pandas",
        "tables",
        "requests",
        "PyYAML",
        "SimpleITK",
        "spekpy",
        "icrp107-database",
        "loguru",
    ]
    + install_requires_windows,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

### Derenzo phantom with 6 sets of cylinders of different size for each set
### For phantom1 set of cylinders each cylinder is assigned with an activity value
### The same procedure as in phantom1 with the activity can be followed to all phantom sets

import opengate as gate
import opengate.contrib.phantoms.derenzo as derenzo
import itk
from opengate.tests import utility

# Define the units used in the simulation set-up
mm = gate.g4_units.mm
cm = gate.g4_units.cm
Bq = gate.g4_units.Bq
cm3 = gate.g4_units.cm3
MeV = gate.g4_units.MeV
BqmL = Bq / cm3

if __name__ == "__main__":
    # create the simulation
    sim = gate.Simulation()

    paths = utility.get_default_test_paths(__file__, "", "test093")

    # Add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # Change world size
    world = sim.world
    world.size = [100 * cm, 100 * cm, 100 * cm]

    # Add the derenzo phantom
    derenzo_phantom = derenzo.add_derenzo_phantom(sim)
    a = 20 * BqmL
    activity_Bq_mL = [10 * a, 2 * a, 3 * a, 4 * a, 5 * a, 6 * a]
    sources = derenzo.add_sources(sim, derenzo_phantom, activity_Bq_mL)
    for source in sources:
        source.particle = "alpha"
        source.energy.type = "mono"
        source.energy.mono = 100 * MeV
        source.direction.type = "momentum"
        source.direction.momentum = [0, 1, 0]

    # Add a plan above the phantom
    plan = sim.add_volume("Box", "plan")
    plan.mother = world.name
    plan.size = [35 * cm, 1 * cm, 35 * cm]
    plan.material = "Tungsten"
    plan.translation = derenzo_phantom.translation
    plan.translation[1] += 10 * cm

    # Physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.enable_decay = True

    # Acquisition options
    sim.check_volumes_overlap = True
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.visu_type = "qt"
    # sim.visu_verbose = True
    # sim.progress_bar = True
    sim.number_of_threads = 1
    sim.output_dir = paths.output
    sim.random_seed = 23654987

    # Actor
    stats_actor = sim.add_actor("SimulationStatisticsActor", "stats")
    stats_actor.track_types_flag = True
    stats_actor.output_filename = "test093_derenzo_stats.txt"

    dose_actor = sim.add_actor("DoseActor", "dose")
    dose_actor.edep.output_filename = "test093_derenzo_edep.mhd"
    dose_actor.attached_to = plan.name
    dose_actor.size = [350, 1, 350]
    dose_actor.spacing = [1 * mm, 10 * mm, 1 * mm]

    # run
    sim.run()

    # compare stats
    stats_ref = utility.read_stat_file(paths.output_ref / "test093_derenzo_stats.txt")
    is_ok = utility.assert_stats(stats_actor, stats_ref, tolerance=0.05)
    is_ok = is_ok and utility.assert_img_sum(
        itk.imread(paths.output_ref / "test093_derenzo_edep.mhd"),
        itk.imread(paths.output / "test093_derenzo_edep.mhd"),
        sum_tolerance=5,
    )
    utility.test_ok(is_ok)



from scipy.spatial.transform import Rotation
from box import Box
import opengate as gate
from opengate.utility import g4_units, get_contrib_path
from opengate.geometry.utility import get_grid_repetition
from opengate.geometry.volumes import unite_volumes, intersect_volumes, subtract_volumes
from opengate.geometry import volumes
import scipy
import numpy as np
import itk


def add_linac(sim, linac_name, sad=1000):
    # materials
    add_linac_materials(sim)

    # linac
    linac = add_empty_linac_box(sim, linac_name, sad)

    # add elements
    add_target(sim, linac.name)
    add_primary_collimator(sim, linac.name)
    add_flattening_filter(sim, linac.name)
    add_ionizing_chamber(sim, linac.name)
    add_back_scatter_plate(sim, linac.name)
    add_mirror(sim, linac.name)

    # kill actor above the target
    kill_around_target(sim, linac.name)

    return linac


def add_empty_linac_box(sim, linac_name, sad=1000):
    # units
    m = g4_units.m
    mm = g4_units.mm
    linac = sim.add_volume("Box", linac_name)
    linac.material = "G4_AIR"
    linac.size = [1 * m, 1 * m, 0.52 * m]
    translation_linac_box = np.array([0 * mm, 0, sad - linac.size[2] / 2 + 3.5 * mm])
    # Isocenter begin at the end of the target, That's why 3.5 mm is added.
    linac.translation = translation_linac_box
    linac.color = [1, 1, 1, 0]
    return linac


def add_linac_materials(sim):
    contrib_paths = get_contrib_path() / "linacs"
    file = contrib_paths / "elekta_versa_materials.db"
    sim.volume_manager.add_material_database(str(file))


def add_target(sim, linac_name):
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    # unit
    mm = g4_units.mm
    nm = g4_units.nm

    # colors
    red = [1, 0.2, 0.2, 0.8]
    green = [0, 1, 0, 0.2]

    # material
    target_material = f"target_tungsten"
    copper = f"target_copper"

    # target
    target_support = sim.add_volume("Tubs", f"{linac_name}_target_support")
    target_support.mother = linac_name
    target_support.material = "G4_AIR"
    target_support.rmin = 0
    target_support.rmax = 15 * mm
    target_support.dz = 13.5 * mm / 2.0
    target_support.translation = [0, 0, z_linac / 2 - target_support.dz - 1 * nm]
    target_support.color = [0, 1, 0, 1]

    target = sim.add_volume("Tubs", f"{linac_name}_target")
    target.mother = target_support.name
    target.material = target_material
    target.rmin = 0
    target.rmax = 2.7 * mm
    target.dz = 1 * mm / 2.0
    target.translation = [0, 0, target_support.dz - 3 * mm]
    target.color = red

    target_support_top = sim.add_volume("Tubs", f"{linac_name}_target_support_top")
    target_support_top.mother = target_support.name
    target_support_top.material = copper
    target_support_top.rmin = 2.7 * mm
    target_support_top.rmax = 15 * mm
    target_support_top.dz = 3.5 * mm / 2.0
    target_support_top.translation = [0, 0, target_support.dz - 1.75 * mm]
    target_support_top.color = green

    target_support_bottom = sim.add_volume(
        "Tubs", f"{linac_name}_target_support_bottom"
    )
    target_support_bottom.mother = target_support.name

    target_support_bottom.material = copper
    target_support_bottom.rmin = 0
    target_support_bottom.rmax = 15 * mm
    target_support_bottom.dz = 10 * mm / 2.0
    target_support_bottom.translation = [0, 0, target_support.dz - 8.5 * mm]
    target_support_bottom.color = green

    return target_support


def kill_around_target(sim, linac_name):
    mm = g4_units.mm
    nm = g4_units.nm
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]

    # above the target
    target_above = sim.add_volume("Tubs", f"target_kill_volume1")
    target_above.mother = linac_name
    target_above.material = "G4_AIR"
    target_above.rmin = 0
    target_above.rmax = 15 * mm
    target_above.dz = 0.5 * nm
    target_above.translation = [0, 0, z_linac / 2 - 0.5 * nm]
    target_above.color = [1, 0, 0, 1]

    # around the target
    target_around = sim.add_volume("Tubs", f"target_kill_volume2")
    target_around.mother = linac_name
    target_around.material = "G4_AIR"
    target_around.rmin = 15.1 * mm
    target_around.rmax = target_around.rmin + 1 * nm
    target_around.dz = 13.5 / 2 * mm
    target_around.translation = [0, 0, z_linac / 2 - target_around.dz - 1 * nm]
    target_around.color = [1, 0, 0, 1]

    # psycho killer
    killer = sim.add_actor("KillActor", f"target_kill")
    killer.attached_to = [target_above.name, target_around.name]


def add_primary_collimator(sim, linac_name):
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    mm = g4_units.mm
    deg = g4_units.deg
    primary_collimator = sim.add_volume("Cons", f"{linac_name}_primary_collimator")
    primary_collimator.mother = linac_name
    primary_collimator.material = "mat_leaf"
    primary_collimator.rmin1 = 31.45 * mm
    primary_collimator.rmax1 = 82 * mm
    primary_collimator.rmin2 = 6.45 * mm
    primary_collimator.rmax2 = 82 * mm
    primary_collimator.dz = 101 * mm / 2.0
    primary_collimator.sphi = 0
    primary_collimator.dphi = 360 * deg
    primary_collimator.translation = [0, 0, z_linac / 2 - 70.6 * mm]
    primary_collimator.color = [0, 0, 1, 0.8]


def add_flattening_filter(sim, linac_name):
    # unit
    mm = g4_units.mm
    deg = g4_units.deg

    # colors
    yellow = [0, 0.7, 0.7, 0.8]
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]

    # bounding cylinder
    flattening_filter = sim.add_volume("Tubs", f"{linac_name}_flattening_filter")
    flattening_filter.mother = linac_name
    flattening_filter.material = "G4_AIR"
    flattening_filter.rmin = 0
    flattening_filter.rmax = 54 * mm
    flattening_filter.dz = 26.1 * mm / 2
    flattening_filter.translation = [0, 0, z_linac / 2 - 149.55 * mm]
    flattening_filter.color = [1, 0, 0, 0]  # invisible

    # create all cones
    def add_cone(sim, p):
        c = sim.add_volume("Cons", f"{linac_name}_flattening_filter_cone_{p.name}")
        c.mother = flattening_filter.name
        c.material = "flattening_filter_material_stain_steel"
        c.rmin1 = 0
        c.rmax1 = p.rmax1
        c.rmin2 = 0
        c.rmax2 = p.rmax2
        c.dz = p.dz
        c.sphi = 0
        c.dphi = 360 * deg
        c.translation = [0, 0, p.tr]
        c.color = yellow

    cones = np.array(
        [
            [0.001, 2.6, 1.2, flattening_filter.dz - 0.6],
            [2.6, 9, 4.9, flattening_filter.dz - 3.65],
            [9, 14.5, 4.9, flattening_filter.dz - 8.55],
            [14.5, 22.5, 5.5, flattening_filter.dz - 13.75],
            [22.5, 32.5, 5.6, flattening_filter.dz - 19.3],
        ]
    )

    cyl = volumes.TubsVolume(name=f"{linac_name}_flattening_filter_cyl_base")
    cyl.rmin = 0
    cyl.rmax = 46.5 * mm
    cyl.dz = 4 * mm

    trap = volumes.TrapVolume(name=f"{linac_name}_flattening_filter_trap_base")
    dz = 93 / 2 * mm
    dy1 = 171.39 / 2 * mm
    dy2 = 36.26 * mm / 2

    dx1 = 4 / 2 * mm
    alpha1 = 0
    alpha2 = alpha1
    phi = 0
    theta = 0
    dx2 = dx1
    dx3 = dx1
    dx4 = dx1

    trap.dx1 = dx1
    trap.dx2 = dx2
    trap.dx3 = dx3
    trap.dx4 = dx4
    trap.dy1 = dy1
    trap.dy2 = dy2
    trap.dz = dz
    trap.alp1 = alpha1
    trap.alp2 = alpha2
    trap.theta = theta
    trap.phi = phi
    rot = Rotation.from_euler("Y", -90, degrees=True).as_matrix()
    ff_base = intersect_volumes(cyl, trap, [0, 0, 0], rot)
    ff_base.name = f"{linac_name}_flattening_filter_base"
    sim.volume_manager.add_volume(ff_base, f"{linac_name}_flattening_filter_base")
    ff_base.mother = flattening_filter.name
    ff_base.translation = [0, 0, flattening_filter.dz - 24.1 * mm]
    ff_base.color = yellow
    ff_base.material = "flattening_filter_material_stain_steel"

    cons_1 = volumes.ConsVolume(name=f"{linac_name}_flattening_filter_cons_1")
    cons_1.rmin1 = 77 / 2 * mm
    cons_1.rmax1 = 93 / 2 * mm
    cons_1.rmin2 = 77 / 2 * mm
    cons_1.rmax2 = 93 / 2 * mm
    cons_1.dz = 3.5 * mm / 2.0
    cons_1.sphi = 0
    cons_1.dphi = 360 * deg

    ff_cons_1 = intersect_volumes(cons_1, trap, [0, 0, 0], rot)
    ff_cons_1.name = f"{linac_name}_flattening_filter_cons_1"
    sim.volume_manager.add_volume(ff_cons_1, f"{linac_name}_flattening_filter_cons_1")
    ff_cons_1.mother = flattening_filter.name
    ff_cons_1.material = "flattening_filter_material_stain_steel"
    ff_cons_1.translation = [0, 0, flattening_filter.dz - 20.35 * mm]
    ff_cons_1.color = yellow

    ff_cons_2 = sim.add_volume("ConsVolume", f"{linac_name}_flattening_filter_cons_2")
    ff_cons_2.rmin1 = 77 / 2 * mm
    ff_cons_2.rmax1 = 79.99 / 2 * mm
    ff_cons_2.rmin2 = 77 / 2 * mm
    ff_cons_2.rmax2 = 79.99 / 2 * mm
    ff_cons_2.dz = 3.5 * mm / 2.0
    ff_cons_2.sphi = 0
    ff_cons_2.dphi = 360 * deg
    ff_cons_2.mother = flattening_filter.name
    ff_cons_2.material = "flattening_filter_material_stain_steel"
    ff_cons_2.translation = [0, 0, flattening_filter.dz - 16.85 * mm]
    ff_cons_2.color = yellow

    ff_cons_3 = sim.add_volume("ConsVolume", f"{linac_name}_flattening_filter_cons_3")
    ff_cons_3.rmin1 = 73 / 2 * mm
    ff_cons_3.rmax1 = 77 / 2 * mm
    ff_cons_3.rmin2 = 73 / 2 * mm
    ff_cons_3.rmax2 = 77 / 2 * mm
    ff_cons_3.dz = 22.1 * mm / 2.0
    ff_cons_3.sphi = 0
    ff_cons_3.dphi = 360 * deg
    ff_cons_3.mother = flattening_filter.name
    ff_cons_3.material = "flattening_filter_material_stain_steel"
    ff_cons_3.translation = [0, 0, flattening_filter.dz - 11.05 * mm]
    ff_cons_3.color = yellow

    ff_cons_4 = sim.add_volume("ConsVolume", f"{linac_name}_flattening_filter_cons_4")
    ff_cons_4.rmin1 = 77 / 2 * mm
    ff_cons_4.rmax1 = 79 / 2 * mm
    ff_cons_4.rmin2 = 77 / 2 * mm
    ff_cons_4.rmax2 = 79 / 2 * mm
    ff_cons_4.dz = 15.1 * mm / 2.0
    ff_cons_4.sphi = 0
    ff_cons_4.dphi = 360 * deg
    ff_cons_4.mother = flattening_filter.name
    ff_cons_4.material = "flattening_filter_material_stain_steel"
    ff_cons_4.translation = [0, 0, flattening_filter.dz - 7.55 * mm]
    ff_cons_4.color = yellow

    ## cons 5 is for the moment an assumption of the secondary filter carrier design
    ff_cons_5 = sim.add_volume("ConsVolume", f"{linac_name}_flattening_filter_cons_5")
    ff_cons_5.rmin1 = 79.99 / 2 * mm
    ff_cons_5.rmax1 = 106 / 2 * mm
    ff_cons_5.rmin2 = 79.99 / 2 * mm
    ff_cons_5.rmax2 = 106 / 2 * mm
    ff_cons_5.dz = 5 * mm / 2.0
    ff_cons_5.sphi = 0
    ff_cons_5.dphi = 360 * deg
    ff_cons_5.mother = flattening_filter.name
    ff_cons_5.material = "flattening_filter_material_mild_steel"
    ff_cons_5.translation = [0, 0, flattening_filter.dz - 16.1 * mm]
    ff_cons_5.color = yellow

    i = 0
    for c in cones:
        cone = Box()
        cone.name = i
        cone.rmax2 = c[0] * mm
        cone.rmax1 = c[1] * mm
        cone.dz = c[2] * mm / 2  # /2 to keep same values than Gate (where dz was /2)
        cone.tr = c[3] * mm
        add_cone(sim, cone)
        i = i + 1


def add_ionizing_chamber(sim, linac_name):
    # unit
    mm = g4_units.mm

    # main cylinder
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    ionizing_chamber = sim.add_volume("Tubs", f"{linac_name}_ionizing_chamber")
    ionizing_chamber.mother = linac_name
    ionizing_chamber.material = "G4_AIR"
    ionizing_chamber.rmin = 0
    ionizing_chamber.rmax = 45 * mm
    ionizing_chamber.dz = 9.28 * mm / 2
    ionizing_chamber.translation = [0, 0, z_linac / 2 - 172.5 * mm]
    ionizing_chamber.color = [0, 0, 0, 0]

    # layers
    def add_layer(sim, p):
        l = sim.add_volume("Tubs", f"{linac_name}_ionizing_chamber_mylar_layer_{p.i}")
        l.mother = ionizing_chamber.name
        l.material = "linac_mylar"
        l.rmin = 0
        l.rmax = 45 * mm
        l.dz = 0.012 * mm / 2
        l.translation = [0, 0, p.tr1]

        l = sim.add_volume("Tubs", f"{linac_name}_ionizing_chamber_carbon_layer_{p.i}")
        l.mother = ionizing_chamber.name
        l.material = "linac_carbon"
        l.rmin = 0
        l.rmax = 45 * mm
        l.dz = 0.000150 * mm / 2
        l.translation = [0, 0, p.tr2]

    layers = [
        [-2.634, -2.627925],
        [-0.434, -0.427925],
        [0.566, 0.572075],
        [1.566, 1.572075],
        [2.566, 2.572075],
        [3.566, 3.572075],
    ]
    i = 1
    for l in layers:
        ll = Box()
        ll.i = i
        ll.tr1 = l[0] * mm
        ll.tr2 = l[1] * mm
        add_layer(sim, ll)
        i = i + 1


def add_back_scatter_plate(sim, linac_name):
    # back_scatter_plate
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    mm = g4_units.mm
    bsp = sim.add_volume("Box", f"{linac_name}_back_scatter_plate")
    bsp.mother = linac_name
    bsp.material = "linac_aluminium"
    bsp.size = [116 * mm, 84 * mm, 3 * mm]
    bsp.translation = [0, 0, z_linac / 2 - 187.5 * mm]
    bsp.color = [1, 0.7, 0.7, 0.8]


def add_mirror(sim, linac_name):
    # unit
    mm = g4_units.mm
    blue = [0, 0, 1, 0.8]

    # main box
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    m = sim.add_volume("Box", f"{linac_name}_mirror")
    m.mother = linac_name
    m.material = "G4_AIR"
    m.size = [137 * mm, 137 * mm, 1.5 * mm]
    m.translation = [0, 0, z_linac / 2 - 227.5 * mm]
    rot = Rotation.from_euler("x", 37.5, degrees=True)
    m.rotation = rot.as_matrix()

    # mylar
    l = sim.add_volume("Box", f"{linac_name}_mirror_mylar_layer")
    l.mother = m.name
    l.material = "linac_mylar"
    l.size = [110 * mm, 110 * mm, 0.0012 * mm]
    l.translation = [0, 0, 0.15 * mm]
    l.color = blue

    # alu
    l = sim.add_volume("Box", f"{linac_name}_mirror_alu_layer")
    l.mother = m.name
    l.material = "linac_aluminium"
    l.size = [110 * mm, 110 * mm, 0.0003 * mm]
    l.translation = [0, 0, -0.6 * mm]
    l.color = blue


def enable_brem_splitting(sim, linac_name, splitting_factor):
    # create a region
    linac_target = sim.volume_manager.get_volume(f"{linac_name}_target")
    region_linac = sim.physics_manager.add_region(name=f"{linac_target.name}_region")
    region_linac.associate_volume(linac_target)
    # set the brem splitting
    s = f"/process/em/setSecBiasing eBrem {region_linac.name} {splitting_factor} 50 MeV"
    sim.g4_commands_after_init.append(s)


def add_electron_source(sim, linac_name, ekin, sx, sy):
    MeV = g4_units.MeV
    mm = g4_units.mm
    nm = g4_units.nm
    deg = g4_units.deg
    source = sim.add_source("GenericSource", f"{linac_name}_e-_source")
    source.particle = "e-"
    source.attached_to = f"{linac_name}_target"
    source.energy.type = "gauss"
    source.energy.mono = ekin * MeV
    source.energy.sigma_gauss = source.energy.mono * (0.08 / 2.35)
    source.position.type = "disc"
    source.position.sigma_x = sx * mm
    source.position.sigma_y = sy * mm
    source.position.translation = [0, 0, 0.5 * mm - 1 * nm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, -1]
    source.direction_relative_to_attached_volume = True
    source.n = 10
    return source


def set_cut_on_linac_head(sim, linac_name, particles, cuts):
    ##FIXME Probably not adapted to the current geometry of the LINAC, need to be corrected
    l_volume = [
        f"{linac_name}_mlc",
        "jaw_box_right",
        "jaw_box_left",
        f"{linac_name}_primary_collimator",
        f"{linac_name}_flattening_filter",
        f"{linac_name}_back_scatter_plate",
        f"{linac_name}_target_support_top",
        f"{linac_name}_target_support_bottom",
    ]
    mm = g4_units.mm
    reg = sim.physics_manager.add_region(name=f"{linac_name}_region")
    for volume in l_volume:
        reg.associate_volume(volume)
    if type(particles) == str:
        if particles == "all":
            reg.production_cuts.gamma = cuts
            reg.production_cuts.electron = cuts
            reg.production_cuts.positron = cuts
    else:
        for i, particle in enumerate(particles):
            if particle == "gamma":
                reg.production_cuts.gamma = cuts[i]
            elif particle == "electron":
                reg.production_cuts.electron = cuts[i]
            elif particle == "positron":
                reg.production_cuts.positron = cuts[i]


def add_phase_space_plane(sim, linac_name, src_phsp_distance):
    mm = g4_units.mm
    m = g4_units.m
    nm = g4_units.nm
    plane = sim.add_volume("Tubs", f"{linac_name}_phsp_plane")
    plane.mother = linac_name
    plane.material = "G4_AIR"
    plane.rmin = 0
    plane.rmax = 0.5 * m / 2 - 0.1 * mm
    plane.dz = 1 * nm  # half height
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    plane.translation = [0 * mm, 0 * mm, +z_linac / 2 - src_phsp_distance]
    plane.color = [1, 0, 0, 1]  # red
    return plane


def add_phase_space_actor(sim, plane_name, i=0):
    if i == 0:
        phsp = sim.add_actor("PhaseSpaceActor", f"{plane_name}_phsp")
    else:
        phsp = sim.add_actor("PhaseSpaceActor", f"{plane_name}_phsp_" + str(i))
    phsp.attached_to = plane_name
    phsp.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "PreDirection",
        "PreDirectionLocal",
        "PDGCode",
    ]
    return phsp


def add_phase_space_source(sim, plane_name):
    source = sim.add_source("PhaseSpaceSource", "phsp_source_global")
    source.attached_to = plane_name
    source.position_key = "PrePositionLocal"
    source.direction_key = "PreDirectionLocal"
    # source.weight_key = "Weight"
    source.global_flag = False
    source.particle = ""
    source.batch_size = 100000
    # source.translate_position = True
    return source


def mlc_leaf(linac_name):
    mm = g4_units.mm
    interleaf_gap = 0.09 * mm
    leaf_length = 155 * mm
    leaf_height = 90 * mm
    leaf_mean_width = 1.76 * mm
    tongues_length = 0.8 * mm

    cyl = volumes.TubsVolume(name=f"{linac_name}_cylinder_leaf")
    cyl.rmin = 0
    cyl.rmax = 170 * mm

    box_rot_leaf = volumes.BoxVolume(name=f"{linac_name}_Box_leaf")
    box_rot_leaf.size = [200 * mm, leaf_length, leaf_height]

    trap_leaf = volumes.TrapVolume(name=f"{linac_name}_trap_leaf")
    dz = leaf_height / 2
    dy1 = leaf_length / 2
    dx1 = 1.94 * mm / 2
    dx3 = 1.58 * mm / 2
    theta = 0

    alpha1 = 0
    alpha2 = alpha1
    phi = 0
    dy2 = dy1
    dx2 = dx1
    dx4 = dx3

    trap_leaf.dx1 = dx1
    trap_leaf.dx2 = dx2
    trap_leaf.dx3 = dx3
    trap_leaf.dx4 = dx4
    trap_leaf.dy1 = dy1
    trap_leaf.dy2 = dy2
    trap_leaf.dz = dz
    trap_leaf.alp1 = alpha1
    trap_leaf.alp2 = alpha2
    trap_leaf.theta = theta
    trap_leaf.phi = phi

    rot_leaf = Rotation.from_euler("Z", -90, degrees=True).as_matrix()
    rot_cyl = Rotation.from_euler("X", 90, degrees=True).as_matrix()

    trap_tongue = volumes.TrapVolume(name=f"{linac_name}_trap_tongue")
    dz = tongues_length / 2
    dy1 = leaf_length / 2

    ##FIXME I need to remove 2 um to the tongues to avoid an overleap between leaves
    dx1 = (interleaf_gap - 2.2 * 10 ** (-3) * mm) / 2
    dx3 = dx1
    alpha1 = 0
    alpha2 = alpha1
    theta = np.arctan((1.58 * mm - 1.94 * mm) * 0.5 / leaf_height)
    phi = 0
    dy2 = dy1
    dx2 = dx1
    dx4 = dx1

    trap_tongue.dx1 = dx1
    trap_tongue.dx2 = dx2
    trap_tongue.dx3 = dx3
    trap_tongue.dx4 = dx4
    trap_tongue.dy1 = dy1
    trap_tongue.dy2 = dy2
    trap_tongue.dz = dz
    trap_tongue.alp1 = alpha1
    trap_tongue.alp2 = alpha2
    trap_tongue.theta = theta
    trap_tongue.phi = phi

    bool_leaf = intersect_volumes(box_rot_leaf, trap_leaf, [0, 0, 0], rot_leaf)
    bool_tongue = intersect_volumes(box_rot_leaf, trap_tongue, [0, 0, 0], rot_leaf)
    bool_leaf = unite_volumes(
        bool_leaf, bool_tongue, [0 * mm, (leaf_mean_width + interleaf_gap) / 2, 0 * mm]
    )
    bool_leaf = intersect_volumes(bool_leaf, cyl, [92.5 * mm, 0, 7.5 * mm], rot_cyl)

    return bool_leaf


def add_mlc(sim, linac_name):

    mm = g4_units.mm
    cm = g4_units.cm
    linac = sim.volume_manager.get_volume(linac_name)
    leaf_height = 90 * mm
    z_linac = linac.size[2]
    center_mlc = 349.3 * mm + 3.5 * mm
    interleaf_gap = 0.09 * mm
    leaf_width = 1.76 * mm
    leaf_lenght = 155 * mm
    nb_leaf = 160
    rotation_angle = np.arctan((1.94 * mm - 1.58 * mm) * 0.5 / leaf_height)

    mlc = sim.add_volume("Box", f"{linac_name}_mlc")
    mlc_bank_rotation = Rotation.from_euler(
        "X", np.arctan(3.25 / 349.3), degrees=False
    ).as_matrix()
    mlc.rotation = mlc_bank_rotation
    mlc.size = [linac.size[0] - 2 * cm, linac.size[1] - 2 * cm, 100 * mm]
    mlc.translation = np.array([0, 0, z_linac / 2 - center_mlc])
    mlc.mother = linac_name
    mlc.color = [0, 0, 0, 0]

    leaf_color = [0.92, 0.61, 0.32, 0.4]
    leaf = mlc_leaf(linac_name)
    leaf.name = f"{linac_name}_leaf"
    sim.volume_manager.add_volume(leaf, f"{linac_name}_leaf")
    leaf.material = "mat_leaf"
    leaf.mother = mlc
    leaf.color = leaf_color

    size = [2, int(0.5 * nb_leaf), 1]
    tr_blocks = np.array([leaf_lenght, leaf_width + interleaf_gap, 0])
    leaves_pos = np.array(get_grid_repetition(size, tr_blocks))
    l_rotation = []
    l_center_translation = []

    for i in range(len(leaves_pos)):
        if i <= int(nb_leaf / 4) - 1:
            angle = 2 * rotation_angle * (i - int(nb_leaf / 4)) + rotation_angle
        elif int(nb_leaf / 4) - 1 < i <= 2 * int(nb_leaf / 4) - 1:
            angle = 2 * rotation_angle * (i - int(nb_leaf / 4) + 1) - rotation_angle
        elif 2 * int(nb_leaf / 4) - 1 < i <= 3 * int(nb_leaf / 4) - 1:
            angle = 2 * rotation_angle * (i - 3 * int(nb_leaf / 4)) + rotation_angle
        elif 3 * int(nb_leaf / 4) - 1 < i:
            angle = 2 * rotation_angle * (i - 3 * int(nb_leaf / 4) + 1) - rotation_angle

        cos_angle = np.cos(np.abs(angle))
        sin_angle = np.sin(np.abs(angle))
        tan_angle_pos = np.tan(np.pi / 2 - rotation_angle - np.abs(angle))
        tan_angle_neg = np.tan(np.pi / 2 - rotation_angle + np.abs(angle))

        left = (leaf_width / 2) * (cos_angle + sin_angle / tan_angle_pos)
        right = (leaf_width / 2) * (cos_angle - sin_angle / tan_angle_neg)

        if angle < 0:
            l_center_translation.append([left, right])
        else:
            l_center_translation.append([right, left])

        if leaves_pos[i, 0] > 0:
            rot = Rotation.from_euler("X", angle, degrees=False).as_matrix()
            l_rotation.append(rot)
        else:
            rot = Rotation.from_euler("XZ", [angle, np.pi], degrees=False).as_matrix()
            l_rotation.append(rot)

    translation_to_apply = []
    l_center_translation = l_center_translation[: int(len(l_center_translation) / 2)]

    for i in range(len(l_center_translation)):
        if i != 0:
            translation_to_apply.append(
                l_center_translation[i - 1][1] + l_center_translation[i][0]
            )

    translation_to_apply = np.array(translation_to_apply) - leaf_width
    translation_to_apply = translation_to_apply[
        : int(len(translation_to_apply) / 2) + 1
    ]
    translation_to_apply[-1] = translation_to_apply[-1] / 2
    final_translation_to_apply = []
    for j in range(len(translation_to_apply)):
        final_translation_to_apply.append(np.sum(translation_to_apply[j:]))

    final_translation_to_apply = np.array(
        (
            (
                (-np.array(final_translation_to_apply)).tolist()
                + final_translation_to_apply[::-1]
            )
            * 2
        )
    )
    for i in range(len(leaves_pos)):
        leaves_pos[i, 1] += final_translation_to_apply[i]
    leaf.translation = leaves_pos
    leaf.rotation = np.array(l_rotation)

    return leaf


def trap_g4_param(
    obj, dx1, dx2, dx3, dx4, dy1, dy2, dz, theta=0, phi=0, alpha1=0, alpha2=0
):
    obj.dx1 = dx1
    obj.dx2 = dx2
    obj.dx3 = dx3
    obj.dx4 = dx4
    obj.dy1 = dy1
    obj.dy2 = dy2
    obj.dz = dz
    obj.alp1 = alpha1
    obj.alp2 = alpha2
    obj.theta = theta
    obj.phi = phi


def add_jaws(sim, linac_name):
    return [add_jaw(sim, linac_name, "left"), add_jaw(sim, linac_name, "right")]


def add_jaw(sim, linac_name, side):
    nm = g4_units.nm
    mm = g4_units.mm
    um = g4_units.um
    cm = g4_units.cm
    jaw_x = 210 * mm
    jaw_y = 220 * mm
    jaw_z = 77 * mm
    linac = sim.volume_manager.get_volume(linac_name)
    center_jaw_z = 470.5 * mm

    ## Traps definition for three trapzoid composing the front f&ace of the jaw in the order of G4 paramaters : dx1,dx2,dx3,dx4,dy1,dy2,dz
    traps_y_dimensions = [
        [32 * mm, 32 * mm, 32 * mm],
        [194 * mm, 200 * mm, 210 * mm],
        [32 * mm, 32 * mm, 32 * mm],
        [194 * mm, 200 * mm, 210 * mm],
        [84.822 * mm, 87.964 * mm, 93.2 * mm],
        [84.822 * mm, 87.964 * mm, 93.2 * mm],
        [10 * mm, 29 * mm, 16 * mm],
    ]

    traps_y_positions = [
        [0, 0, 0],
        [
            jaw_y / 2 - 16.3 * mm - (93.2 - 84.822) * mm - (84.822) / 2 * mm,
            jaw_y / 2 - 16.3 * mm - (93.2 - 87.964) * mm - (87.964) / 2 * mm,
            jaw_y / 2 - 16.3 * mm - 93.2 / 2 * mm,
        ],
        [
            jaw_z / 2 - 22 * mm - 10 * mm / 2,
            jaw_z / 2 - 32 * mm - 29 * mm / 2,
            jaw_z / 2 - 32 * mm - 29 * mm - 16 * mm / 2,
        ],
    ]

    boxes_to_complete_traps_y_dimensions = [
        [194 * mm, 200 * mm, 210 * mm],
        [(109.5 - 84.822) * mm, (109.5 - 87.964) * mm, 16.3 * mm],
        [10 * mm, 29 * mm, 16 * mm],
    ]
    boxes_to_complete_traps_y_positions = [
        [0, 0, 0],
        [
            jaw_y / 2 - (109.5 - 84.822) / 2 * mm,
            jaw_y / 2 - (109.5 - 87.964) / 2 * mm,
            jaw_y / 2 - 16.3 / 2 * mm,
        ],
        [
            jaw_z / 2 - 22 * mm - 10 * mm / 2,
            jaw_z / 2 - 32 * mm - 29 * mm / 2,
            jaw_z / 2 - 32 * mm - 29 * mm - 16 * mm / 2,
        ],
    ]
    middle_trap_dimension = [
        [32 * mm, 32 * mm],
        [73.2 * mm, 73.2 * mm],
        [45 * mm, 45 * mm],
    ]
    middle_trap_position = [
        0,
        jaw_y / 2 - 109.5 * mm - 73.2 * mm / 2,
        jaw_z / 2 - 32 * mm - 45 * mm / 2,
    ]

    jaw_box = sim.volume_manager.add_volume("Box", "jaw_box_" + side)
    jaw_box.mother = linac_name
    jaw_box.size = np.array([jaw_x, jaw_y, jaw_z])
    jaw_box.translation = [0, 0, linac.size[2] / 2 - center_jaw_z - 3.5 * mm]
    jaw_box.color = [0, 0, 0, 0]

    middle_trap = sim.volume_manager.add_volume("Trap", "middle_trap_" + side)
    trap_g4_param(
        middle_trap,
        middle_trap_dimension[0][1] / 2,
        middle_trap_dimension[0][1] / 2,
        middle_trap_dimension[0][0] / 2,
        middle_trap_dimension[0][0] / 2,
        middle_trap_dimension[1][0] / 2,
        middle_trap_dimension[1][1] / 2,
        middle_trap_dimension[2][0] / 2,
    )
    middle_trap.translation = middle_trap_position
    middle_trap.color = [0.8, 0.4, 0.65, 0.6]
    middle_trap.mother = jaw_box.name
    middle_trap.material = "mat_leaf"

    for i in range(1, len(traps_y_positions)):
        traps = sim.volume_manager.add_volume("Trap", "trap_" + side + "_" + str(i))
        trap_g4_param(
            traps,
            traps_y_dimensions[0][i] / 2,
            traps_y_dimensions[1][i] / 2,
            traps_y_dimensions[2][i] / 2,
            traps_y_dimensions[3][i] / 2,
            traps_y_dimensions[4][i] / 2,
            traps_y_dimensions[5][i] / 2,
            traps_y_dimensions[6][i] / 2,
        )

        traps.translation = [
            traps_y_positions[0][i],
            traps_y_positions[1][i],
            traps_y_positions[2][i],
        ]
        traps.mother = jaw_box.name
        traps.material = "mat_leaf"
        traps.color = [0.8, 0.4, 0.65, 0.6]

    # Jaws curve tips
    cylindre = volumes.TubsVolume(name=f"{linac_name}_cyl_leaf" + "_" + side)
    cylindre.rmin = 0
    cylindre.rmax = 135 * mm
    cylindre.dz = jaw_x
    rot_cyl = Rotation.from_euler("Y", 90, degrees=True).as_matrix()

    for i in range(3):
        box = volumes.BoxVolume(name="Box_to_complete_traps_y_" + side + "_" + str(i))
        box.size = [
            boxes_to_complete_traps_y_dimensions[0][i],
            boxes_to_complete_traps_y_dimensions[1][i],
            boxes_to_complete_traps_y_dimensions[2][i],
        ]
        box.translation = [
            boxes_to_complete_traps_y_positions[0][i],
            boxes_to_complete_traps_y_positions[1][i],
            boxes_to_complete_traps_y_positions[2][i],
        ]
        box.mother = jaw_box.name
        the_box = intersect_volumes(
            box,
            cylindre,
            [
                0,
                boxes_to_complete_traps_y_dimensions[1][i] / 2 - 135 * mm,
                jaw_z / 2 - 35 * mm - boxes_to_complete_traps_y_positions[2][i],
            ],
            rot_cyl,
        )
        sim.volume_manager.add_volume(the_box, "curved_" + box.name)
        the_box.mother = jaw_box.name
        the_box.material = "mat_leaf"
        the_box.color = [0.8, 0.4, 0.65, 0.6]

    top_jaw_box = volumes.BoxVolume(name="top_jaw_box_" + side)
    top_jaw_box.size = [194 * mm, jaw_y, 22 * mm]
    top_jaw_box.translation = [0, 0, jaw_z / 2 - 22 * mm / 2]

    top_jaw_box = intersect_volumes(
        top_jaw_box,
        cylindre,
        [
            0,
            top_jaw_box.size[1] / 2 - 135 * mm,
            jaw_z / 2 - 35 * mm - top_jaw_box.translation[2],
        ],
        rot_cyl,
    )
    dimension_corner = np.array(
        [
            np.sqrt(22**2 + 6.9**2) * mm,
            np.sqrt(22**2 + 6.9**2) * 6.9 / 22 * mm,
            22 * mm + 10 * um,
        ]
    )

    for i in range(2):
        corner_to_remove = volumes.BoxVolume(
            name="corner_to_remove_" + side + "_" + str(i)
        )
        corner_to_remove.size = dimension_corner
        angle_for_corner = np.arctan(6.9 / 22) * 180 / np.pi
        if i == 1:
            angle_for_corner = -angle_for_corner
        rot_for_corner = Rotation.from_euler("z", -angle_for_corner, degrees=True)
        translation_to_apply = np.array(
            [
                194 * mm / 2 - corner_to_remove.size[0] / 2,
                -jaw_y / 2 - corner_to_remove.size[1] / 2,
                0,
            ]
        )
        t = gate.geometry.utility.get_translation_from_rotation_with_center(
            rot_for_corner,
            [+corner_to_remove.size[0] / 2, -corner_to_remove.size[1] / 2, 0],
        )
        if i == 1:
            translation_to_apply[0] = -translation_to_apply[0]
            t = gate.geometry.utility.get_translation_from_rotation_with_center(
                rot_for_corner,
                [-corner_to_remove.size[0] / 2, -corner_to_remove.size[1] / 2, 0],
            )
        translation_to_apply += t
        top_jaw_box = subtract_volumes(
            top_jaw_box,
            corner_to_remove,
            translation_to_apply,
            rot_for_corner.as_matrix(),
        )
    sim.volume_manager.add_volume(top_jaw_box, "curved_" + top_jaw_box.name)
    top_jaw_box.material = "mat_leaf"
    top_jaw_box.mother = jaw_box.name
    top_jaw_box.color = [0.8, 0.4, 0.65, 0.6]

    below_top_jaw_box = volumes.BoxVolume(name="below_top_jaw_box_" + side)
    below_top_jaw_box.size = [154 * mm, jaw_y - (109.5 - 84.822) * mm, 10 * mm]
    below_top_jaw_box_size = list(below_top_jaw_box.size)
    below_top_jaw_box.translation = [
        0,
        jaw_y / 2 - (109.5 - 84.822) * mm - (jaw_y - (109.5 - 84.822)) / 2 * mm,
        jaw_z / 2 - 22 * mm - 10 * mm / 2,
    ]
    dimension_corner = np.array(
        [
            np.sqrt(22**2 + 6.9**2) * mm,
            np.sqrt(22**2 + 6.9**2) * 6.9 / 22 * mm,
            10 * mm + 10 * um,
        ]
    )
    for i in range(2):
        corner_to_remove = volumes.BoxVolume(
            name="corner_to_remove_" + side + "_" + str(i + 2)
        )
        corner_to_remove.size = dimension_corner
        angle_for_corner = np.arctan(6.9 / 22) * 180 / np.pi
        if i == 1:
            angle_for_corner = -angle_for_corner
        rot_for_corner = Rotation.from_euler("z", -angle_for_corner, degrees=True)
        translation_to_apply = np.array(
            [
                194 * mm / 2 - corner_to_remove.size[0] / 2,
                -(jaw_y - (109.5 - 84.822) * mm) / 2 - corner_to_remove.size[1] / 2,
                0,
            ]
        )
        t = gate.geometry.utility.get_translation_from_rotation_with_center(
            rot_for_corner,
            [+corner_to_remove.size[0] / 2, -corner_to_remove.size[1] / 2, 0],
        )
        if i == 1:
            translation_to_apply[0] = -translation_to_apply[0]
            t = gate.geometry.utility.get_translation_from_rotation_with_center(
                rot_for_corner,
                [-corner_to_remove.size[0] / 2, -corner_to_remove.size[1] / 2, 0],
            )
        translation_to_apply += t
        below_top_jaw_box = subtract_volumes(
            below_top_jaw_box,
            corner_to_remove,
            translation_to_apply,
            rot_for_corner.as_matrix(),
        )
    left_edge_box_to_remove = volumes.BoxVolume(name="left_edge_box_to_remove_" + side)
    left_edge_box_to_remove.size = [12.05 * mm, 59 * mm, 10 * mm + 10 * um]
    size_box = left_edge_box_to_remove.size
    # [- 154 * mm / 2 + size_box[0] / 2, below_top_jaw_box_size[1] / 2 - 89 * mm - size_box[1] / 2,
    #  jaw_z / 2 - 22 * mm - 10 * mm / 2]
    below_top_jaw_box = subtract_volumes(
        below_top_jaw_box,
        left_edge_box_to_remove,
        [
            -154 * mm / 2 + size_box[0] / 2,
            -below_top_jaw_box_size[1] / 2 + 80 * mm + size_box[1] / 2,
            0,
        ],
        np.identity(3),
    )
    sim.volume_manager.add_volume(
        below_top_jaw_box, "truncated_" + below_top_jaw_box.name
    )
    below_top_jaw_box.material = "mat_leaf"
    below_top_jaw_box.mother = jaw_box.name
    below_top_jaw_box.color = [0.8, 0.4, 0.65, 0.6]

    for i in range(2):
        trap_to_complete_jaw = sim.volume_manager.add_volume(
            "Trap", "trap_to_complete_jaw_" + side + "_" + str(i)
        )
        alpha = np.arctan(0.5 * 20 / 20.8775)
        if i == 1:
            alpha = -alpha
        trap_g4_param(
            trap_to_complete_jaw,
            1 * nm / 2,
            20 * mm / 2,
            1 * nm / 2,
            20 * mm / 2,
            20.8775 / 2 * mm,
            20.8775 / 2 * mm,
            10 * mm / 2,
            alpha1=alpha,
            alpha2=alpha,
        )

        trap_to_complete_jaw.color = [0.8, 0.4, 0.65, 0.6]
        trap_to_complete_jaw.mother = jaw_box.name
        if i == 0:
            trap_to_complete_jaw.translation = [
                jaw_x / 2 - (210 - 194) / 2 * mm - 20 / 2 * mm - 10 / 2 * mm + 1 * nm,
                jaw_y / 2 - (109.5 - 84.822) * mm - 20.8775 / 2 * mm,
                jaw_z / 2 - 22 * mm - 10 * mm / 2,
            ]
        if i == 1:
            trap_to_complete_jaw.translation = [
                -(
                    jaw_x / 2
                    - (210 - 194) / 2 * mm
                    - 20 / 2 * mm
                    - 10 / 2 * mm
                    + 1 * nm
                ),
                jaw_y / 2 - (109.5 - 84.822) * mm - 20.8775 / 2 * mm,
                jaw_z / 2 - 22 * mm - 10 * mm / 2,
            ]
        trap_to_complete_jaw.color = [0.8, 0.4, 0.65, 0.6]
        trap_to_complete_jaw.material = "mat_leaf"

    for i in range(2):
        traps_edge_jaw = sim.volume_manager.add_volume(
            "Trap", "traps_edge_jaw_" + side + "_" + str(i)
        )
        alpha = -np.arctan(0.5 * 6.324 / 10.64)
        if i == 1:
            alpha = -alpha

        # alpha = 0
        trap_g4_param(
            traps_edge_jaw,
            (1 * nm) / 2,
            (57.5 * mm) / 2,
            (1 * nm) / 2,
            (57.5 * mm) / 2,
            21.28 * mm / 2 * mm,
            21.28 * mm / 2 * mm,
            15 * mm / 2,
            alpha1=alpha,
            alpha2=alpha,
        )
        traps_edge_jaw.color = [0.8, 0.4, 0.65, 0.6]
        traps_edge_jaw.mother = jaw_box.name
        traps_edge_jaw.translation = np.array(
            [
                jaw_x / 2
                - (210 - 194) / 2 * mm
                - 20 * mm
                - 0.5 * 21.28 * mm * np.tan(alpha)
                - 57.5 * mm / 2,
                jaw_y / 2 - (109.5 - 84.822) * mm - 20.8775 * mm - 21.28 * mm / 2,
                jaw_z / 2 - 32 * mm - 15 * mm / 2,
            ]
        )
        if i == 1:
            traps_edge_jaw.translation[0] = (
                -jaw_x / 2
                + (210 - 194) / 2 * mm
                + 20 * mm
                - 0.5 * 21.28 * mm * np.tan(alpha)
                + 57.5 * mm / 2
            )
        rot_angle = np.arctan(87.964 / ((200 - 32) / 2))
        if i == 1:
            rot_angle = -rot_angle
        rot = Rotation.from_euler("z", rot_angle * 180 / np.pi, degrees=True)
        rot_inversion = Rotation.from_euler(
            "xz", [180, -180 + rot_angle * 180 / np.pi], degrees=True
        )
        vec_translation_for_rotation = np.array(
            [57.5 * mm / 2 + 0.5 * 21.28 * mm * np.tan(alpha), 21.28 * mm / 2, 0]
        )
        if i == 1:
            vec_translation_for_rotation[0] = (
                -57.5 * mm / 2 + 0.5 * 21.28 * mm * np.tan(alpha)
            )
        t = gate.geometry.utility.get_translation_from_rotation_with_center(
            rot, vec_translation_for_rotation
        )
        traps_edge_jaw.translation += t
        translation_for_overlap_correction = np.array([32.19 * um, -32.19 * um, 0])
        if i == 1:
            translation_for_overlap_correction[0] = -translation_for_overlap_correction[
                0
            ]
        traps_edge_jaw.translation += np.array(translation_for_overlap_correction)
        if i == 0:
            traps_edge_jaw.rotation = rot.as_matrix()
        else:
            traps_edge_jaw.rotation = rot_inversion.as_matrix()
        traps_edge_jaw.material = "mat_leaf"

    band_box_to_add_left_side = sim.volume_manager.add_volume(
        "Box", "band_box_to_add_left_side_" + side
    )
    band_box_to_add_left_side.size = [20 * mm, 10 * mm, 10 * mm]
    band_box_to_add_left_side.translation = [
        -194 * mm / 2 + band_box_to_add_left_side.size[0] / 2,
        -jaw_y / 2 + 70 * mm + band_box_to_add_left_side.size[1] / 2,
        jaw_z / 2 - 22 * mm - 10 * mm / 2,
    ]
    band_box_to_add_left_side.mother = jaw_box.name
    band_box_to_add_left_side.color = [0.8, 0.4, 0.65, 0.6]
    band_box_to_add_left_side.material = "mat_leaf"

    if side == "left":
        jaw_box.translation += np.array([0, -jaw_y / 2, 0])
    if side == "right":
        rot_jaw = Rotation.from_euler("Z", 180, degrees=True).as_matrix()
        jaw_box.translation += np.array([0, jaw_y / 2, 0])
        jaw_box.rotation = rot_jaw

    return jaw_box


def retrieve_offset_data_to_apply(
    offset_type: object, collimation_type: object
) -> object:
    mlc_offset_thales_to_radiation_field = [
        [
            -149.19,
            -144.56,
            -135.25,
            -125.90,
            -116.51,
            -107.06,
            -97.57,
            -88.03,
            -78.45,
            -68.81,
            -59.12,
            -49.39,
            -39.61,
            -29.78,
            -19.9,
            -9.98,
            0,
            10.02,
            20.10,
            30.22,
            40.39,
            50.61,
            60.88,
            71.19,
            81.55,
            91.97,
            102.43,
            112.94,
            123.49,
            134.1,
            144.75,
            155.44,
            166.19,
            176.98,
            187.92,
            198.71,
            209.64,
        ],
        [
            -5.53,
            -5.17,
            -4.47,
            -3.82,
            -3.22,
            -2.67,
            -2.16,
            -1.70,
            -1.29,
            -0.93,
            -0.61,
            -0.35,
            -0.13,
            0.04,
            0.16,
            0.23,
            0.25,
            0.23,
            0.15,
            0.03,
            -0.14,
            -0.36,
            -0.63,
            -0.94,
            -1.31,
            -1.72,
            -2.18,
            -2.69,
            -3.25,
            -3.85,
            -4.51,
            -5.20,
            -5.95,
            -6.74,
            -7.58,
            -8.47,
            -9.4,
        ],
    ]

    jaws_offset_thales_to_radiation_field = [
        [
            -127.57,
            -117.93,
            -108.26,
            -98.56,
            -88.83,
            -79.08,
            -69.29,
            -59.48,
            -49.64,
            -39.77,
            -29.87,
            -19.94,
            -9.99,
            -0,
            10.01,
            20.06,
            30.13,
            40.23,
            50.36,
            60.52,
            70.71,
            80.92,
            91.17,
            101.44,
            111.74,
            122.07,
            132.43,
            142.82,
            153.23,
            163.68,
            174.15,
            184.64,
            195.17,
            205.72,
        ],
        [
            -2.18,
            -1.82,
            -1.49,
            -1.19,
            -0.92,
            -0.68,
            -0.46,
            -0.28,
            -0.12,
            0.01,
            0.11,
            0.18,
            0.23,
            0.24,
            0.22,
            0.18,
            0.11,
            0.01,
            -0.13,
            -0.28,
            -0.47,
            -0.69,
            -0.93,
            -1.21,
            -1.51,
            -1.84,
            -2.20,
            -2.59,
            -3,
            -3.44,
            -3.91,
            -4.41,
            -4.94,
            -5.49,
        ],
    ]

    mlc_offset_light_field_to_radiation_field = [
        [
            -149.19,
            -144.56,
            -135.25,
            -125.90,
            -116.51,
            -107.06,
            -97.57,
            -88.03,
            -78.45,
            -68.81,
            -59.12,
            -49.39,
            -39.61,
            -29.78,
            -19.9,
            -9.98,
            0,
            10.02,
            20.10,
            30.22,
            40.39,
            50.61,
            60.88,
            71.19,
            81.55,
            91.97,
            102.43,
            112.94,
            123.49,
            134.1,
            144.75,
            155.44,
            166.19,
            176.98,
            187.92,
            198.71,
            209.64,
        ],
        [
            0.28,
            0.28,
            0.28,
            0.27,
            0.27,
            0.27,
            0.27,
            0.27,
            0.27,
            0.26,
            0.26,
            0.26,
            0.26,
            0.26,
            0.26,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
        ],
    ]

    jaws_offset_light_field_to_radiation_field = [
        [
            -127.57,
            -117.93,
            -108.26,
            -98.56,
            -88.83,
            -79.08,
            -69.29,
            -59.48,
            -49.64,
            -39.77,
            -29.87,
            -19.94,
            -9.99,
            -0,
            10.01,
            20.06,
            30.13,
            40.23,
            50.36,
            60.52,
            70.71,
            80.92,
            91.17,
            101.44,
            111.74,
            122.07,
            132.43,
            142.82,
            153.23,
            163.68,
            174.15,
            184.64,
            195.17,
            205.72,
        ],
        [
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.24,
            0.23,
            0.23,
            0.23,
            0.23,
            0.23,
            0.23,
            0.23,
            0.23,
            0.23,
            0.23,
            0.23,
            0.23,
        ],
    ]

    if collimation_type == "jaw":
        if offset_type == "from thales":
            return (
                jaws_offset_thales_to_radiation_field[0],
                jaws_offset_thales_to_radiation_field[1],
            )
        elif offset_type == "from light field":
            return (
                jaws_offset_light_field_to_radiation_field[0],
                jaws_offset_light_field_to_radiation_field[1],
            )

    elif collimation_type == "mlc":
        if offset_type == "from thales":
            return (
                mlc_offset_thales_to_radiation_field[0],
                mlc_offset_thales_to_radiation_field[1],
            )
        elif offset_type == "from light field":
            return (
                mlc_offset_light_field_to_radiation_field[0],
                mlc_offset_thales_to_radiation_field[1],
            )


def define_pos_jaws_rectangular_field(y_field, sad=1000):
    mm = g4_units.mm
    center_jaws = 470.5 * mm
    jaws_height = 77 * mm
    center_curve_jaws = center_jaws - (jaws_height / 2 - 35 * mm)

    jaws_y_aperture = y_field / 2 * center_curve_jaws / sad
    pos_y_jaws = np.array(
        [-int(10 * jaws_y_aperture) / 10, int(10 * jaws_y_aperture) / 10]
    )
    return pos_y_jaws


def define_pos_mlc_jaws_rectangular_field(
    mlc, x_field, y_field, sad=1000, opposite_leaf_gap=0
):
    mm = g4_units.mm

    center_mlc = 349.3 * mm
    center_jaws = 470.5 * mm
    jaws_height = 77 * mm
    center_curve_mlc = center_mlc - 7.5 * mm
    leaf_width = 1.76 * mm + 0.09 * mm
    center_curve_jaws = center_jaws - (jaws_height / 2 - 35 * mm)

    #### way to correct the thales correction at the isocenter of an offset to obtain the light field ####
    #### Not used here since we directly have the offset to apply from thales to obtain the radiation field

    # R_mlc = 170 * mm
    # R_jaws = 135 * mm
    # theta_x = np.arctan(0.5 * x_field / sad)
    # theta_y = np.arctan(0.5 * y_field / sad)

    # mlc_x_aperture = (x_field / 2) * (center_curve_mlc + R_mlc * np.sin(theta_x)) / sad - R_mlc * (1 - np.cos(theta_x))
    # jaws_y_aperture = (y_field / 2) * (center_curve_jaws + R_jaws * np.sin(theta_y)) / sad - R_jaws * (1 - np.cos(theta_y))
    ###############################################

    mlc_position, mlc_offset = retrieve_offset_data_to_apply("from thales", "mlc")
    jaw_position, jaw_offset = retrieve_offset_data_to_apply("from thales", "jaw")

    f_offset_mlc = scipy.interpolate.interp1d(mlc_position, mlc_offset, kind="cubic")
    f_offset_jaw = scipy.interpolate.interp1d(jaw_position, jaw_offset, kind="cubic")

    mlc_x_aperture = (center_curve_mlc / sad) * (
        x_field / 2 - f_offset_mlc(x_field / 2)
    )
    jaws_y_aperture = (center_curve_jaws / sad) * (
        y_field / 2 - f_offset_jaw(y_field / 2)
    )

    max_field_jaw = np.sqrt(110**2 - (jaws_height / 2) ** 2)
    max_y_aperture = (110 * mm - max_field_jaw) + jaws_y_aperture
    y_position_for_opening = max_y_aperture * 349.3 / (470.5 - (jaws_height / 2))

    leaves_position_y = mlc.translation[:, 1]
    leaves_ID = np.arange(160)
    leaves_to_open = leaves_ID[
        (leaves_position_y > -(y_position_for_opening + leaf_width / 2))
        & (leaves_position_y < y_position_for_opening + leaf_width / 2)
    ]
    leaves_to_open_left = leaves_to_open[leaves_to_open < 80]
    leaves_to_open_right = leaves_to_open[leaves_to_open >= 80]

    opposite_leaf_gap = opposite_leaf_gap * center_mlc / sad
    pos_x_leaves = np.zeros(160)
    pos_x_leaves[0:80] -= opposite_leaf_gap / 2
    pos_x_leaves[80:160] += opposite_leaf_gap / 2
    pos_x_leaves[leaves_to_open_left] = -mlc_x_aperture
    pos_x_leaves[leaves_to_open_right] = mlc_x_aperture
    pos_x_leaves = np.array(10 * pos_x_leaves, dtype=int) / 10
    pos_y_jaws = np.array(
        [-int(10 * jaws_y_aperture) / 10, int(10 * jaws_y_aperture) / 10]
    )

    return pos_x_leaves, pos_y_jaws


def field(mlc, jaws, pos_x_leaves, pos_y_jaws):
    mlc.translation[:, 0] += pos_x_leaves
    for i, jaw in enumerate(jaws):
        jaw.translation[1] += pos_y_jaws[i]


def convert_degree_minuts_to_decimal_degree(degree, minuts):
    decimal_degree = degree + 1 / 60 * minuts
    return decimal_degree


def set_rectangular_field(mlc, jaws, x_field, y_field, sad=1000, opposite_leaf_gap=0):
    pos_x_leaves, pos_y_jaws = define_pos_mlc_jaws_rectangular_field(
        mlc, x_field, y_field, sad, opposite_leaf_gap
    )

    field(mlc, jaws, pos_x_leaves, pos_y_jaws)


def linac_rotation(sim, linac_name, angles, cp_id="all_cp"):
    gantry_angle = angles[0]
    collimation_angle = angles[1]
    linac = sim.volume_manager.get_volume(linac_name)
    rotations = []
    translations = []
    if cp_id == "all_cp":
        nb_cp_id = len(gantry_angle)
        cp_id = np.arange(0, nb_cp_id, 1)
    translation_linac = linac.translation
    for n in cp_id:
        rot = Rotation.from_euler(
            "YZ", [gantry_angle[n], collimation_angle[n]], degrees=True
        )
        t = gate.geometry.utility.get_translation_from_rotation_with_center(
            rot, [0, 0, -translation_linac[2]]
        )
        rot = rot.as_matrix()
        rotations.append(rot)
        translations.append(np.array(t) + translation_linac)
    linac.add_dynamic_parametrisation(translation=translations, rotation=rotations)


def translation_from_sad(sim, linac_name, translation, sad=1000):
    linac = sim.volume_manager.get_volume(linac_name)
    linac.translation = np.array(translation)
    linac.translation[2] += sad - linac.size[2] / 2


def rotation_around_user_point(
    sim, linac_name, str_axes, angle_list, point_coordinate=[0, 0, 0]
):
    point_coordinate = np.array(point_coordinate)
    linac = sim.volume_manager.get_volume(linac_name)
    translation = linac.translation
    rot = Rotation.from_euler(str_axes, angle_list, degrees=True)
    new_translation = gate.geometry.utility.get_translation_from_rotation_with_center(
        rot, point_coordinate - np.array(translation)
    )
    linac.translation = translation + new_translation
    linac.rotation = rot.as_matrix()


def jaw_dynamic_translation(jaw, jaw_positions, side, cp_id="all_cp", sad=1000):
    mm = g4_units.mm
    translations = []
    rotations = []
    jaw_height = 77 * mm
    center_jaw = 470.5 * mm
    center_curve_jaw = center_jaw - (jaw_height / 2 - 35 * mm)
    rot_jaw = Rotation.from_euler("Z", 180, degrees=True).as_matrix()
    if cp_id == "all_cp":
        nb_cp_id = len(jaw_positions)
        cp_id = np.arange(0, nb_cp_id, 1)
    for n in cp_id:
        jaw_position, jaw_offset = retrieve_offset_data_to_apply("from thales", "jaw")
        f_offset_jaw = scipy.interpolate.interp1d(
            jaw_position, jaw_offset, kind="cubic"
        )
        if side == "left":
            jaw_pos = -jaw_positions[n]
        if side == "right":
            jaw_pos = jaw_positions[n]
        jaws_y_aperture = (center_curve_jaw / sad) * (jaw_pos - f_offset_jaw(jaw_pos))
        if side == "left":
            jaws_y_aperture = -jaws_y_aperture
            rotations.append(np.identity(3))
        if side == "right":
            rotations.append(rot_jaw)
        jaw_translation = jaw.translation + np.array([0, jaws_y_aperture, 0])
        translations.append(jaw_translation)
    jaw.add_dynamic_parametrisation(translation=translations, rotation=rotations)


def mlc_leaves_dynamic_translation(
    mlc, leaves_position, cp_id="all_cp", sad=1000, only_return_position=False
):
    mm = g4_units.mm
    center_mlc = 349.3 * mm
    center_curve_mlc = center_mlc - 7.5 * mm
    nb_leaves = 160
    motion_leaves_t = []
    motion_leaves_r = []
    mlc_leaves_name = []

    mlc_position, mlc_offset = retrieve_offset_data_to_apply("from thales", "mlc")
    f_offset_mlc = scipy.interpolate.interp1d(mlc_position, mlc_offset, kind="cubic")

    for i in range(nb_leaves):
        mlc_leaves_name.append(mlc.name + "_rep_" + str(i))

    for i in range(nb_leaves):
        motion_leaves_t.append([])
        motion_leaves_r.append([])

    translation_mlc = []
    if cp_id == "all_cp":
        nb_cp_id = len(leaves_position)
        cp_id = np.arange(0, nb_cp_id, 1)
    for n in cp_id:
        mlc_x_aperture_left = (center_curve_mlc / sad) * (
            leaves_position[n] - f_offset_mlc(-leaves_position[n])
        )
        mlc_x_aperture_right = (center_curve_mlc / sad) * (
            leaves_position[n] - f_offset_mlc(leaves_position[n])
        )
        mlc_x_aperture = np.array(
            mlc_x_aperture_left[:80].tolist() + mlc_x_aperture_right[80:].tolist()
        )
        for i in range(len(mlc_x_aperture)):
            translation_mlc.append(np.copy(mlc.translation[i]))
            motion_leaves_t[i].append(
                translation_mlc[i] + np.array([mlc_x_aperture[i], 0, 0])
            )
            motion_leaves_r[i].append(mlc.rotation[i])
    for i in range(nb_leaves):
        mlc.add_dynamic_parametrisation(
            repetition_index=i, translation=motion_leaves_t[i]
        )


def set_linac_head_motion(
    sim, linac_name, jaws, mlc, rt_plan_parameters, cp_id="all_cp", sad=1000
):
    leaves_position = rt_plan_parameters["leaves"]
    jaw_1_positions = rt_plan_parameters["jaws 1"]
    jaw_2_positions = rt_plan_parameters["jaws 2"]
    linac_head_positions = rt_plan_parameters["gantry angle"]
    collimation_rotation = rt_plan_parameters["collimation angle"]
    mlc_leaves_dynamic_translation(mlc, leaves_position, cp_id, sad)
    jaw_dynamic_translation(jaws[0], jaw_1_positions, "left", cp_id, sad)
    jaw_dynamic_translation(jaws[1], jaw_2_positions, "right", cp_id, sad)
    linac_rotation(sim, linac_name, [linac_head_positions, collimation_rotation], cp_id)


def get_patient_translation_and_rotation_from_RT_plan_to_IEC(rt_plan_parameters, img):
    isocenter = rt_plan_parameters["isocenter"][0]
    image = itk.imread(img)
    offset = np.array(image.GetOrigin())
    dim = np.array(image.GetLargestPossibleRegion().GetSize())
    spacing = np.array(image.GetSpacing())

    # IMAGE ROTATION ACCORDING TO IEC 61217

    size = (dim - 1) * spacing
    center = offset + size / 2
    rotation_to_apply = Rotation.from_euler("X", -90, degrees=True).as_matrix()

    isocenter_vector = isocenter - center
    rotated_isocenter_vector = np.dot(rotation_to_apply, isocenter_vector)

    isocenter_rot_img = center + rotated_isocenter_vector
    translation_to_apply = center - isocenter_rot_img

    return (translation_to_apply, rotation_to_apply)


def set_time_intervals_from_rtplan(sim, rt_plan_parameters, cp_id="all_cp"):
    MU = rt_plan_parameters["weight"]
    sec = gate.g4_units.s
    sim.run_timing_intervals = []
    if cp_id == "all_cp":
        nb_cp_id = len(MU)
        cp_id = np.arange(0, nb_cp_id, 1)
    for i in cp_id:
        if i == 0:
            sim.run_timing_intervals.append([0, MU[0] * sec])
        else:
            sim.run_timing_intervals.append(
                [np.sum(MU[:i]) * sec, np.sum(MU[: i + 1]) * sec]
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import os
import opengate as gate
from opengate.tests import utility


def main():
    paths = utility.get_default_test_paths(
        __file__, "gate_test044_pbs_rot_transl", "test044_rot_transl"
    )

    particle = "Carbon_"
    energy = "1440MeV_"
    beam_shape = "sourceShapePBS"
    folder = particle + energy + beam_shape

    output_path = paths.output
    ref_path = paths.gate_output

    # for loop
    start = -500
    spacing = 100
    end = -start + spacing
    planePositionsV = range(start, end, spacing)

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 123654789
    sim.random_engine = "MersenneTwister"
    sim.output_dir = paths.output

    # units
    km = gate.g4_units.km
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    deg = gate.g4_units.deg
    mrad = gate.g4_units.mrad

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    # waterbox
    # translation and rotation like in the Gate macro
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [10 * cm, 10 * cm, 100.2 * cm]
    waterbox.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    waterbox.translation = [-50.3 * cm, 0 * cm, 0 * cm]
    waterbox.material = "Vacuum"
    waterbox.color = [0, 0, 1, 1]

    # Planes
    m = Rotation.identity().as_matrix()

    for i in planePositionsV:
        plane = sim.add_volume("Box", "planeNr" + str(i) + "a")
        plane.mother = "waterbox"
        plane.size = [100 * mm, 100 * mm, 2 * mm]
        plane.translation = [0 * mm, 0 * mm, i * mm]
        plane.rotation = m
        plane.material = "G4_AIR"
        plane.color = [1, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "FTFP_INCLXX_EMZ"
    sim.physics_manager.global_production_cuts.all = 1000 * km

    # default source for tests (from test42)
    source = sim.add_source("IonPencilBeamSource", "mysource")
    source.energy.mono = 1440 * MeV
    # source.energy.type = 'gauss'
    source.particle = "ion 6 12"  # carbon
    source.position.type = "disc"  # pos = Beam, shape = circle + sigma
    # rotate the disc, equiv to : rot1 0 1 0 and rot2 0 0 1
    source.position.rotation = Rotation.from_euler("y", -90, degrees=True).as_matrix()
    source.position.translation = [-100 * mm, 20 * mm, 30 * mm]
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]
    source.n = 20000
    source.direction.partPhSp_x = [
        2.3335754 * mm,
        2.3335754 * mrad,
        0.00078728 * mm * mrad,
        0,
    ]
    source.direction.partPhSp_y = [
        1.96433431 * mm,
        0.00079118 * mrad,
        0.00249161 * mm * mrad,
        0,
    ]

    count = 0
    # add dose actors
    for i in planePositionsV:
        dose = sim.add_actor("DoseActor", "doseInYZ" + str(i))
        filename = "plane" + str(i) + "a.mhd"
        dose.output_filename = filename
        dose.attached_to = "planeNr" + str(i) + "a"
        dose.size = [250, 250, 1]
        dose.spacing = [0.4, 0.4, 2]
        dose.hit_type = "random"
        count += 1

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "Stats")
    s.track_types_flag = True

    print(sim.source_manager.dump_sources())

    # create output dir, if it doesn't exist
    output_path.mkdir(parents=True, exist_ok=True)

    # start simulation
    sim.run()

    # print results at the end
    stat = sim.get_actor("Stats")
    print(stat)

    print("Start to analyze data")
    # override = False
    # if (not os.path.exists(ref_path / "sigma_values.txt")) or override:
    #     sigmasRef, musRef = utility.write_gauss_param_to_file(
    #         ref_path,
    #         planePositionsV,
    #         saveFig=False,
    #         fNamePrefix="plane",
    #         fNameSuffix="a_Carbon_1440MeV_sourceShapePBS-Edep.mhd",
    # )
    override = True
    output_pathV = [
        sim.get_actor("doseInYZ" + str(i)).edep.get_output_path()
        for i in planePositionsV
    ]
    if (not os.path.exists(output_path / "sigma_values.txt")) or override:
        sigmasGam, musGam = utility.write_gauss_param_to_file(
            output_pathV, planePositionsV, saveFig=False
        )
    else:
        print("Some data are already available for analysis")

    # ----------------------------------------------------------------------------------------------------------------
    # tests

    # statistics
    stat_file = "SimulationStatistic_" + folder + ".txt"
    stats_ref = utility.read_stat_file(ref_path / stat_file)
    is_ok = utility.assert_stats(stat, stats_ref, 0.15)

    # energy deposition
    for i in planePositionsV:
        print("\nDifference for EDEP plane " + str(i))
        # mhd_gate = "plane" + str(i) + "a.mhd"
        mhd_gate = sim.get_actor("doseInYZ" + str(i)).edep.get_output_path()
        mhd_ref = "plane" + str(i) + "a_" + folder + "-Edep.mhd"
        is_ok = (
            utility.assert_images(
                ref_path / mhd_ref,
                output_path / mhd_gate,
                stat,
                tolerance=50,
                ignore_value_data2=0,
                axis="x",
                img_threshold=0.1,
                test_sad=False,
            )
            and is_ok
        )

        """
        EdepColorMap = utility.create_2D_Edep_colorMap(output_path / mhd_gate)
        img_name = 'Plane_'+str(i)+'ColorMap.png'
        EdepColorMap.savefig(output_path / img_name)
        plt.close(EdepColorMap)
        """

    # beam shape
    print("Comparing sigma values")
    sigma_file = "sigma_values.txt"
    is_ok = (
        utility.compareGaussParamFromFile(
            output_path / sigma_file,
            ref_path / sigma_file,
            rel_tol=2,
            abs_tol=0.5,
            verb=True,
        )
        and is_ok
    )

    print("Comparing mu values")
    sigma_file = "mu_values.txt"
    is_ok = (
        utility.compareGaussParamFromFile(
            output_path / sigma_file,
            ref_path / sigma_file,
            rel_tol=2,
            abs_tol=0.5,
            verb=True,
        )
        and is_ok
    )

    utility.test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests import utility
from test013_phys_lists_helpers import create_pl_sim


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test013_phys_lists")

    # create simulation
    sim = create_pl_sim()

    # remove ion sources
    sim.source_manager.sources.pop("ion1")
    sim.source_manager.sources.pop("ion2")

    # change physics
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMZ"

    # em parameters
    sim.physics_manager.em_parameters.fluo = True
    sim.physics_manager.em_parameters.auger = True
    sim.physics_manager.em_parameters.auger_cascade = True
    sim.physics_manager.em_parameters.pixe = True
    sim.physics_manager.em_parameters.deexcitation_ignore_cut = True

    sim.physics_manager.em_switches_world.deex = True
    sim.physics_manager.em_switches_world.auger = True
    sim.physics_manager.em_switches_world.pixe = True

    print("Phys list cuts:")
    print(sim.physics_manager.dump_production_cuts())

    # start simulation
    # sim.set_g4_verbose(True)
    # sim.g4_commands_after_init.append("/tracking/verbose 1")
    sim.run()

    stats = sim.get_actor("Stats")

    # Gate mac/main_2.mac
    stats_ref = utility.read_stat_file(paths.gate_output / "stat_2.txt")
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.07)

    utility.test_ok(is_ok)
